"""
Helper routines to compute power spectra.
"""

import math

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact

from ._radial_average import compute_binned_radial_average


def compute_binned_powerspectrum(
    fourier_image_or_volume: (
        Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]
    ),
    radial_frequency_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    pixel_or_voxel_size: float | Float[Array, ""] = 1.0,
    *,
    minimum_frequency: float = 0.0,
    maximum_frequency: float = math.sqrt(2) / 2,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"]]:
    """Compute the power spectrum of an image averaged on a set
    of radial bins.

    !!! warning
        If `radial_frequency_grid` is passed in inverse angstroms, then
        it is necessary to also pass the `pixel_size` argument.

        ```python
        from cryojax.coordinates import make_radial_frequency_grid
        from cryojax.image import rfftn

        image, pixel_size = ...
        fourier_image = rfftn(image)
        radial_frequency_grid_in_angstroms = make_radial_frequency_grid(
            image.shape, pixel_size, outputs_rfftfreqs=True
        )
        powerspectrum, bins = compute_binned_powerspectrum(
            rfftn(image), radial_frequency_grid_in_angstroms, pixel_size
        )
        ```

        This is also true for other fourier statistics in `cryojax.image`,
        such as the `fourier_shell_correlation` function.

    **Arguments:**

    - `fourier_image_or_volume`:
        An image or volume in Fourier space.
    - `radial_frequency_grid`:
        The radial frequency coordinate system of `fourier_image_or_volume`.
    - `pixel_or_voxel_size`:
        The pixel or voxel size of `fourier_image_or_volume`. If
        `radial_frequency_grid` is passed in inverse angstroms,
        this argument must be included.
    - `minimum_frequency`:
        Minimum frequency bin. By default, `0.0`. This is
        not measured in inverse angstroms, even if `radial_frequency_grid`
        is in inverse angstroms.
    - `maximum_frequency`:
        Maximum frequency bin. By default, `math.sqrt(2) / 2`. This is
        not measured in inverse angstroms, even if `radial_frequency_grid`
        is in inverse angstroms.

    **Returns:**

    A tuple of the radially averaged power spectrum and the frequency bins
    over which it is computed.
    """
    # Compute squared amplitudes
    squared_fourier_amplitudes = (
        fourier_image_or_volume * jnp.conjugate(fourier_image_or_volume)
    ).real
    # Compute bins
    frequency_bins = _make_radial_frequency_bins(
        fourier_image_or_volume.shape,
        minimum_frequency,
        maximum_frequency,
        pixel_or_voxel_size,
    )
    # Compute radially averaged power spectrum as a 1D profile
    radially_binned_powerspectrum = compute_binned_radial_average(
        squared_fourier_amplitudes, radial_frequency_grid, frequency_bins
    )

    return radially_binned_powerspectrum, frequency_bins


def compute_fourier_ring_correlation(
    fourier_image_1: Inexact[Array, "y_dim x_dim"],
    fourier_image_2: Inexact[Array, "y_dim x_dim"],
    radial_frequency_grid: Float[Array, "y_dim x_dim"],
    pixel_size: float | Float[Array, ""] = 1.0,
    threshold: float | Float[Array, ""] = 0.5,
    *,
    minimum_frequency: float = 0.0,
    maximum_frequency: float = math.sqrt(2) / 2,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"], Float[Array, ""]]:
    """Compute the fourier ring correlation for two images.

    **Arguments:**

    - `fourier_image_1`:
        An image in fourier space. For example, this can be from the output of
        `cryojax.image.rfftn`.
    - `fourier_image_2`:
        Another image in fourier space. See documentation for `image_1`
        for conventions.
    - `radial_frequency_grid`:
        The radial frequency coordinate system of the images.
    - `pixel_size`:
        The pixel of the images. If `radial_frequency_grid` is passed in
        inverse angstroms, this argument must be included.
    - `threshold`:
        The threshold at which to draw the distinction between input images.
    - `minimum_frequency`:
        Minimum frequency bin. By default, `0.0`. This is
        not measured in inverse angstroms, even if `radial_frequency_grid`
        is in inverse angstroms.
    - `maximum_frequency`:
        Maximum frequency bin. By default, `math.sqrt(2) / 2`. This is
        not measured in inverse angstroms, even if `radial_frequency_grid`
        is in inverse angstroms.

    **Returns:**

    - `frc_curve`:
        The fourier ring correlations as a function of `frequency_bins`.
    - `frequency_bins`:
        The array of frequencies for which we have calculated the
        correlations.
    - `frequency_threshold`:
        The frequency at which the correlation drops below the
        specified threshold.
    """
    frc_curve, frequency_bins, frequency_threshold = _compute_fourier_correlation(
        fourier_image_1,
        fourier_image_2,
        radial_frequency_grid,
        pixel_size,
        threshold=threshold,
        minimum_frequency=minimum_frequency,
        maximum_frequency=maximum_frequency,
    )
    return frc_curve, frequency_bins, frequency_threshold


def compute_fourier_shell_correlation(
    volume_1: Inexact[Array, "z_dim y_dim x_dim"],
    volume_2: Inexact[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: Float[Array, "z_dim y_dim x_dim"],
    voxel_size: float | Float[Array, ""] = 1.0,
    threshold: float | Float[Array, ""] = 0.5,
    *,
    minimum_frequency: float = 0.0,
    maximum_frequency: float = math.sqrt(2) / 2,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"], Float[Array, ""]]:
    """Compute the fourier shell correlation for two voxel maps.

    **Arguments:**

    - `volume_1`:
        A volume in real or fourier space. If it is in fourier space,
        it should be from the output of `cryojax.image.rfftn` (e.g.
        the zero-frequency component should be in the corner). If it
        is in real-space, it cannot be complex-valued.
    - `volume_2`:
        A volume in real or fourier space. See documentation for `volume_1`
        for conventions.
    - `radial_frequency_grid`:
        The radial frequency coordinate system of the volumes.
    - `voxel_size`:
        The voxel size of the volumes. If `radial_frequency_grid` is passed
        in inverse angstroms, this argument must be included.
    - `threshold`:
        The threshold at which to draw the distinction between input maps.
        By default, `threshold = 0.5` for two 'known' volumes according to
        the half-bit criterion. If using half-maps derived from ab initio
        refinements, set `threshold = 0.143` by convention.

    **Returns:**

    - `fsc_curve`:
        The fourier shell correlations as a function of `frequency_bins`.
    - `frequency_bins`:
        The array of frequencies for which we have calculated the
        correlations.
    - `frequency_threshold`:
        The frequencies at which the correlation drops below the
        specified threshold.

    !!! warning

        It is common to obtain a `frequency_threshold` given in inverse angstroms.
        This function achieves this behavior if the `voxel_size` argument is passed
        and the `radial_frequency_grid` argument is given in inverse angstroms. For
        example,

        ```python
        from cryojax.coordinates import make_radial_frequency_grid
        from cryojax.image import rfftn

        volume_1, volume_2, voxel_size = ...
        fourier_volume_1, fourier_volume_2 = rfftn(volume_1), rfftn(volume_2)
        radial_frequency_grid = make_radial_frequency_grid(
            volume_1.shape, voxel_size, outputs_rfftfreqs=True
        )
        fsc_curve, frequency_bins, frequency_threshold = compute_fourier_shell_correlation(
            fourier_volume_1, fourier_volume_2, radial_frequency_grid, voxel_size
        )
        ```

        This behavior can be similarly achieved by leaving out the `voxel_size` of
        the functions `make_radial_frequency_grid` and `compute_fourier_shell_correlation`
        and computing `frequency_bins / voxel_size` and `frequency_threshold / voxel_size`.
    """  # noqa: E501
    fsc_curve, frequency_bins, frequency_threshold = _compute_fourier_correlation(
        volume_1,
        volume_2,
        radial_frequency_grid,
        voxel_size,
        threshold=threshold,
        minimum_frequency=minimum_frequency,
        maximum_frequency=maximum_frequency,
    )
    return fsc_curve, frequency_bins, frequency_threshold


def _compute_fourier_correlation(
    fourier_array_1: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    fourier_array_2: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    radial_frequency_grid: (
        Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]
    ),
    grid_spacing: float | Float[Array, ""],
    threshold: float | Float[Array, ""],
    minimum_frequency: float,
    maximum_frequency: float,
) -> tuple[Float[Array, " n_bins"], Float[Array, " n_bins"], Float[Array, ""]]:
    # Compute FSC/FRC radially averaged 1D profile
    correlation_map = (
        (fourier_array_1 * jnp.conjugate(fourier_array_2))
        / jnp.sqrt(jnp.abs(fourier_array_1) ** 2 * jnp.abs(fourier_array_2) ** 2)
    ).real
    frequency_bins = _make_radial_frequency_bins(
        fourier_array_1.shape, minimum_frequency, maximum_frequency, grid_spacing
    )
    correlation_curve = compute_binned_radial_average(
        correlation_map,
        radial_frequency_grid,
        frequency_bins,
    )
    # Find where FSC/FRC drops below the specified threshold
    # TODO: Add van heel criterion.
    where_below_threshold = jnp.where(
        correlation_curve < threshold, 0, 1
    )  # 0s when below, 1s, when above
    # ... find minimum index where we flip from 0 to 1
    where_is_crossing = jnp.diff(where_below_threshold)
    # ... make an array that has a value of its index when we have a crossing, and a dummy
    # value otherwise
    arr_size = where_is_crossing.size
    arr_indices = jnp.arange(arr_size, dtype=int)
    dummy_index = arr_size + 100
    indices_at_0_to_1_flips = jnp.where(where_is_crossing == -1, arr_indices, dummy_index)
    # ... get minimum of array
    threshold_crossing_index = jnp.amin(indices_at_0_to_1_flips) + 1
    frequency_threshold = frequency_bins[threshold_crossing_index]

    return correlation_curve, frequency_bins, frequency_threshold


def _make_radial_frequency_bins(shape, minimum_frequency, maximum_frequency, pixel_size):
    q_min, q_max = minimum_frequency, maximum_frequency
    q_step = 1.0 / max(*shape)
    n_bins = 1 + int((q_max - q_min) / q_step)
    return jnp.linspace(q_min, q_max, n_bins) / pixel_size



from ..simulator._multislice import (
    AbstractMultisliceIntegrator as AbstractMultisliceIntegrator,
    FFTMultisliceIntegrator as FFTMultisliceIntegrator,
    MultisliceScatteringTheory as MultisliceScatteringTheory,
)
from ..simulator._solvent_2d import (
    AbstractRandomSolvent2D as AbstractRandomSolvent2D,
    GRFSolvent2D as GRFSolvent2D,
    SolventMixturePower as SolventMixturePower,
)
from ..simulator._volume import (
    EwaldSphereExtraction as EwaldSphereExtraction,
)



"""
Filters to apply to images in Fourier space
"""

import abc
import math
from typing import ClassVar
from typing_extensions import override

import jax
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact

from ...jax_util import FloatLike, NDArrayLike
from .._coordinates import make_frequency_grid
from .._edges import resize_with_crop_or_pad
from .._fft import irfftn, rfftn
from .._fourier_statistics import compute_binned_powerspectrum
from .._radial_average import radial_average_to_grid
from ._base_transform import AbstractImageTransform


class AbstractFilter(AbstractImageTransform, strict=True):
    """Base class for computing and applying an image filter."""

    is_real_space: ClassVar[bool] = False

    @abc.abstractmethod
    def get(self) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError

    @override
    def __call__(
        self, image: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]
    ) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
        return image * self.get()


class CustomFilter(AbstractFilter, strict=True):
    """Pass a custom filter as an array."""

    array: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        filter: (
            Inexact[NDArrayLike, "y_dim x_dim"]
            | Inexact[NDArrayLike, "z_dim y_dim x_dim"]
        ),
    ):
        self.array = jnp.asarray(filter)

    @override
    def get(self) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.array


class LowpassFilter(AbstractFilter, strict=True):
    """Apply a low-pass filter to an image or volume, with
    a cosine soft-edge.
    """

    array: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        frequency_grid_in_angstroms_or_pixels: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
        grid_spacing: FloatLike = 1.0,
        frequency_cutoff_fraction: FloatLike = 0.95,
        rolloff_width_fraction: FloatLike = 0.05,
    ):
        """**Arguments:**

        - `frequency_grid_in_angstroms_or_pixels`:
            The frequency grid of the image or volume.
        - `grid_spacing`:
            The pixel or voxel size of `frequency_grid_in_angstroms_or_pixels`.
        - `frequency_cutoff_fraction`:
            The cutoff frequency as a fraction of the Nyquist frequency.
            By default, `0.95`.
        - `rolloff_width_fraction`:
            The rolloff width as a fraction of the Nyquist frequency.
            By default, ``0.05``.
        """
        self.array = _compute_lowpass_filter(
            frequency_grid_in_angstroms_or_pixels,
            jnp.asarray(grid_spacing),
            jnp.asarray(frequency_cutoff_fraction),
            jnp.asarray(rolloff_width_fraction),
        )

    @override
    def get(self) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.array


class HighpassFilter(AbstractFilter, strict=True):
    """Apply a low-pass filter to an image or volume, with
    a cosine soft-edge.
    """

    array: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]

    def __init__(
        self,
        frequency_grid_in_angstroms_or_pixels: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
        grid_spacing: FloatLike = 1.0,
        frequency_cutoff_fraction: FloatLike = 0.95,
        rolloff_width_fraction: FloatLike = 0.05,
    ):
        """**Arguments:**

        - `frequency_grid_in_angstroms_or_pixels`:
            The frequency grid of the image or volume.
        - `grid_spacing`:
            The pixel or voxel size of `frequency_grid_in_angstroms_or_pixels`.
        - `frequency_cutoff_fraction`:
            The cutoff frequency as a fraction of the Nyquist frequency.
            By default, `0.95`.
        - `rolloff_width_fraction`:
            The rolloff width as a fraction of the Nyquist frequency.
            By default, ``0.05``.
        """
        self.array = 1.0 - _compute_lowpass_filter(
            frequency_grid_in_angstroms_or_pixels,
            jnp.asarray(grid_spacing),
            jnp.asarray(frequency_cutoff_fraction),
            jnp.asarray(rolloff_width_fraction),
        )

    @override
    def get(self) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.array


class WhiteningFilter(AbstractFilter, strict=True):
    """Compute a whitening filter from an image. This is taken
    to be the inverse square root of the 2D radially averaged
    power spectrum.

    This implementation follows the cisTEM whitening filter
    algorithm.
    """

    array: Inexact[Array, "y_dim x_dim"]

    def __init__(
        self,
        images: Float[NDArrayLike, "_ _"] | Float[NDArrayLike, "_ _ _"],
        shape: tuple[int, int] | None = None,
        *,
        interpolation_mode: str = "linear",
        outputs_squared: bool = False,
    ):
        """**Arguments:**

        - `images`:
            The image (or stack of images) from which to compute the power spectrum.
        - `shape`:
            The shape of the resulting filter. This downsamples or
            upsamples the filter by cropping or padding in real space.
        - `interpolation_mode`:
            The method of interpolating the binned, radially averaged
            power spectrum onto a 2D grid. Either `nearest` or `linear`.
        - `outputs_squared`:
            If `False`, the whitening filter is the inverse square root of the image
            power. If `True`, the filter is the inverse of the image power.
        """
        images = jnp.expand_dims(images, 0) if images.ndim == 2 else jnp.asarray(images)
        self.array = _compute_whitening_filter(
            images,
            shape,
            interpolation_mode=interpolation_mode,
            outputs_squared=outputs_squared,
        )

    @override
    def get(self) -> Inexact[Array, "y_dim x_dim"]:
        return self.array


def _compute_lowpass_filter(
    frequency_grid: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"],
    grid_spacing: Float[Array, ""],
    cutoff_fraction: Float[Array, ""],
    rolloff_width_fraction: Float[Array, ""],
) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
    k_max = 1.0 / (2.0 * grid_spacing)
    cutoff_radius = cutoff_fraction * k_max
    rolloff_width = rolloff_width_fraction * k_max

    radial_frequency_grid = jnp.linalg.norm(frequency_grid, axis=-1)

    def compute_filter_at_frequency(radial_frequency):
        return jnp.where(
            radial_frequency <= cutoff_radius,
            1.0,
            jnp.where(
                radial_frequency > cutoff_radius + rolloff_width,
                0.0,
                0.5
                * (
                    1
                    + jnp.cos(jnp.pi * (radial_frequency - cutoff_radius) / rolloff_width)
                ),
            ),
        )

    compute_filter = (
        jax.vmap(jax.vmap(compute_filter_at_frequency))
        if radial_frequency_grid.ndim == 2
        else jax.vmap(jax.vmap(jax.vmap(compute_filter_at_frequency)))
    )

    return compute_filter(radial_frequency_grid)


def _compute_whitening_filter(
    image_stack: Float[Array, "n_images y_dim x_dim"],
    shape: tuple[int, int] | None,
    *,
    interpolation_mode: str,
    outputs_squared: bool,
) -> Float[Array, "{shape[0]} {shape[1]}"]:
    # Make coordinates
    frequency_grid = make_frequency_grid(image_stack.shape[1:])
    # Transform to fourier space
    n_pixels = math.prod(image_stack.shape[1:])
    fourier_image_stack = rfftn(image_stack, axes=(1, 2)) / jnp.sqrt(n_pixels)
    # Compute norms
    radial_frequency_grid = jnp.linalg.norm(frequency_grid, axis=-1)
    # Compute stack of power spectra
    compute_powerspectrum_stack = jax.vmap(
        lambda im, freq: compute_binned_powerspectrum(
            im, freq, maximum_frequency=math.sqrt(2) / 2
        ),
        in_axes=[0, None],
        out_axes=(0, None),
    )
    binned_powerspectrum_stack, frequency_bins = compute_powerspectrum_stack(
        fourier_image_stack, radial_frequency_grid
    )
    # Take the mean over the stack
    binned_powerspectrum = jnp.mean(binned_powerspectrum_stack, axis=0)
    # Put onto a grid
    binned_powerspectrum_on_grid = radial_average_to_grid(
        binned_powerspectrum,
        frequency_bins,
        radial_frequency_grid,
        interpolation_mode=interpolation_mode,
    )
    # Resize to be the desired shape
    if shape is not None:
        new_shape = shape
        binned_powerspectrum_on_grid = rfftn(
            resize_with_crop_or_pad(
                irfftn(binned_powerspectrum_on_grid, s=image_stack.shape[1:]),
                shape,
                mode="edge",
            )
        ).real
        # ... resizing and going back to fourier space can introduce negative values
        binned_powerspectrum_on_grid = jnp.where(
            binned_powerspectrum_on_grid < 0,
            0.0,
            binned_powerspectrum_on_grid,
        )
    else:
        new_shape = image_stack.shape[1], image_stack.shape[2]
    # Compute inverse square root (or inverse square)
    inverse_fn = jax.lax.reciprocal if outputs_squared else jax.lax.rsqrt
    whitening_filter = jnp.where(
        jnp.isclose(binned_powerspectrum_on_grid, 0.0),
        0.0,
        inverse_fn(binned_powerspectrum_on_grid),
    )
    # Set zero mode to 0, defining the filter to zero out these modes
    whitening_filter = whitening_filter.at[0, 0].set(0.0)
    # If the image size is even, there can be an issue with the nyquist corner
    if new_shape[0] % 2 == 0 and new_shape[1] % 2 == 0:
        whitening_filter = whitening_filter.at[new_shape[0] // 2, new_shape[1] // 2].set(
            0.0
        )

    return whitening_filter



"""Unit conversions."""

import jax.numpy as jnp
import scipy
from jaxtyping import Array, Float

from ..jax_util import FloatLike


def wavelength_from_kilovolts(voltage_in_kilovolts: FloatLike) -> Float[Array, ""]:
    """Get the relativistic electron wavelength at a given accelerating voltage. For
    reference, see Equation 2.5 in Section 2.1 from *Spence, John CH. High-resolution
    electron microscopy. OUP Oxford, 2013.*.

    **Arguments:**

    - `voltage_in_kilovolts`:
        The accelerating voltage given in kilovolts.

    **Returns:**

    The relativistically corrected electron wavelength in Angstroms corresponding to the
    energy `energy_in_keV`.
    """
    accelerating_voltage = 1000.0 * voltage_in_kilovolts  # keV to eV
    return jnp.asarray(
        12.2639 / (accelerating_voltage + 0.97845e-6 * accelerating_voltage**2) ** 0.5,
        dtype=float,
    )


def lorentz_factor_from_kilovolts(voltage_in_kilovolts: FloatLike) -> Float[Array, ""]:
    """Get the Lorentz factor given an accelerating voltage.

    **Arguments:**

    - `voltage_in_kilovolts`:
        The accelerating voltage given in kilovolts.

    **Returns:**

    The Lorentz factor.
    """
    c = scipy.constants.speed_of_light
    m0 = scipy.constants.electron_mass
    e = scipy.constants.elementary_charge
    rest_energy = m0 * c**2
    accelerating_energy = e * (1000.0 * voltage_in_kilovolts)
    return jnp.asarray(1 + accelerating_energy / rest_energy, dtype=float)


def interaction_constant_from_kilovolts(
    voltage_in_kilovolts: FloatLike,
) -> Float[Array, ""]:
    """Get the electron interaction constant given an accelerating voltage.

    The interaction constant is necessary to compute the object
    phase shift distribution from an electrostatic potential integrated
    on the plane.

    !!! info
        In the projection approximation in cryo-EM, the phase shifts in the
        exit plane are given by

        $$\\eta(x, y) = \\sigma_e \\int dz \\ V(x, y, z),$$

        where $\\sigma_e$ is typically referred to as the interaction
        constant. However, in `cryojax`, the potential is rescaled
        to units of inverse length squared as

        $$U(x, y, z) = \\frac{m_0 e}{2 \\pi \\hbar^2} V(x, y, z).$$

        With this rescaling of the potential, the defined as with the
        equation

        $$\\eta(x, y) = \\sigma_e \\int dz \\ U(x, y, z)$$

        with

        $$\\sigma_e = \\lambda \\gamma,$$

        where $\\lambda$ the relativistic electron wavelength $\\gamma$ is
        the lorentz factor.

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - For the definition of the phase shifts in terms of the rescaled potential, see
        Chapter 69, Page 2012, Equation 69.34b from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*

    See the documentation on atom-based scattering potentials for more information.

    **Arguments:**

    - `voltage_in_kilovolts`:
        The accelerating voltage given in kilovolts.

    **Returns:**

    The electron interaction constant.
    """
    wavelength = wavelength_from_kilovolts(voltage_in_kilovolts)
    lorentz_factor = lorentz_factor_from_kilovolts(voltage_in_kilovolts)
    return wavelength * lorentz_factor



import abc
from typing import Generic, Literal, TypeVar
from typing_extensions import Self, override

import equinox as eqx
from jaxtyping import Array, Complex, Float, Inexact, PRNGKeyArray

from ...jax_util import NDArrayLike
from .._image_config import AbstractImageConfig
from .._pose import AbstractPose


VolRep = TypeVar("VolRep", bound="AbstractVolumeRepresentation")
T = TypeVar("T")


ProjectionOrEwaldSphereArray = (
    Complex[
        Array,
        "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
    ]
    | Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
    | Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
)
ProjectionArray = (
    Complex[
        Array,
        "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
    ]
    | Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
)
EwaldSphereArray = (
    Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
    | Float[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]
)

VoxelArray = (
    Inexact[Array, "{self.shape[0]} {self.shape[1]} {self.shape[2]}"]
    | Complex[Array, "{self.shape[0]} {self.shape[1]} {self.shape[2]}//2+1"]
)


class AbstractVolumeParametrization(eqx.Module, strict=True):
    """Abstract interface for a parametrization of a volume. Specifically,
    the cryo-EM image formation process typically starts with a *scattering potential*.
    "Volumes" and "scattering potentials" in cryoJAX are synonymous.

    !!! info
        In, `cryojax`, potentials should be built in units of *inverse length squared*,
        $[L]^{-2}$. This rescaled potential is defined to be

        $$U(\\mathbf{r}) = \\frac{m_0 e}{2 \\pi \\hbar^2} V(\\mathbf{r}),$$

        where $V$ is the electrostatic potential energy, $\\mathbf{r}$ is a positional
        coordinate, $m_0$ is the electron rest mass, and $e$ is the electron charge.

        For a single atom, this rescaled potential has the advantage that under usual
        scattering approximations (i.e. the first-born approximation), the
        fourier transform of this quantity is closely related to tabulated electron scattering
        factors. In particular, for a single atom with scattering factor $f^{(e)}(\\mathbf{q})$
        and scattering vector $\\mathbf{q}$, its rescaled potential is equal to

        $$U(\\mathbf{r}) = \\mathcal{F}^{-1}[f^{(e)}(\\boldsymbol{\\xi} / 2)](\\mathbf{r}),$$

        where $\\boldsymbol{\\xi} = 2 \\mathbf{q}$ is the wave vector coordinate and
        $\\mathcal{F}^{-1}$ is the inverse fourier transform operator in the convention

        $$\\mathcal{F}[f](\\boldsymbol{\\xi}) = \\int d^3\\mathbf{r} \\ \\exp(2\\pi i \\boldsymbol{\\xi}\\cdot\\mathbf{r}) f(\\mathbf{r}).$$

        The rescaled potential $U$ gives the following time-independent schrodinger equation
        for the scattering problem,

        $$(\\nabla^2 + k^2) \\psi(\\mathbf{r}) = - 4 \\pi U(\\mathbf{r}) \\psi(\\mathbf{r}),$$

        where $k$ is the incident wavenumber of the electron beam.

        **References**:

        - For the definition of the rescaled potential, see
        Chapter 69, Page 2003, Equation 69.6 from *Hawkes, Peter W., and Erwin Kasper.
        Principles of Electron Optics, Volume 4: Advanced Wave Optics. Academic Press,
        2022.*
        - To work out the correspondence between the rescaled potential and the electron
        scattering factors, see the supplementary information from *Vulović, Miloš, et al.
        "Image formation modeling in cryo-electron microscopy." Journal of structural
        biology 183.1 (2013): 19-32.*
    """  # noqa: E501

    @abc.abstractmethod
    def to_representation(
        self, rng_key: PRNGKeyArray | None = None
    ) -> "AbstractVolumeRepresentation":
        """Core interface for computing the
        [`cryojax.simulator.AbstractVolumeRepresentation`][] for imaging.

        Users looking to create custom volumes often won't implement this function
        directly, but rather will implement the
        a [`cryojax.simulator.AbstractVolumeRepresentation`][] subclass.
        Implementing a [`cryojax.simulator.AbstractVolumeParametrization`][]
        is useful when there is a distinction between how exactly to parametrize
        the volume for analysis and how to represent it for imaging.

        **Arguments:**

        - `rng_key`:
            An optional RNG key for including noise / stochastic
            elements to volume simulation.
        """
        raise NotImplementedError


class AbstractVolumeRepresentation(AbstractVolumeParametrization, strict=True):
    """Abstract interface for the representation of a volume, such
    as atomic coordinates, voxels, or a neural network.

    Volume representations contain information of coordinates and may be
    passed to [`cryojax.simulator.AbstractVolumeIntegrator`][]
    classes for imaging.
    """

    rotation_convention: eqx.AbstractClassVar[Literal["object", "frame"]]

    @abc.abstractmethod
    def rotate_to_pose(self, pose: AbstractPose, inverse: bool = False) -> Self:
        """Rotate the coordinate system of the volume."""
        raise NotImplementedError

    @override
    def to_representation(self, rng_key: PRNGKeyArray | None = None) -> Self:
        """Since this class is itself an
        `AbstractVolumeRepresentation`, this function maps to the identity.

        **Arguments:**

        - `rng_key`:
            Not used in this implementation.
        """
        del rng_key
        return self


class AbstractAtomVolume(AbstractVolumeRepresentation, strict=True):
    """Abstract interface for a volume represented as a point-cloud."""

    @abc.abstractmethod
    def translate_to_pose(self, pose: AbstractPose) -> Self:
        raise NotImplementedError


class AbstractVoxelVolume(AbstractVolumeRepresentation, strict=True):
    """Abstract interface for a volume represented with voxels.

    !!! info

        If you are using a `volume` in a voxel representation
        pass, the voxel size *must* be passed as the
        `pixel_size` argument, e.g.

        ```python
        import cryojax.simulator as cxs
        from cryojax.io import read_array_from_mrc

        real_voxel_grid, voxel_size = read_array_from_mrc("example.mrc")
        volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)
        ...
        config = cxs.BasicImageConfig(shape, pixel_size=voxel_size, ...)
        ```

        If this is not done, the resulting
        image will be incorrect and *not* rescaled to the specified
        to the different pixel size.
    """

    @property
    @abc.abstractmethod
    def shape(self) -> tuple[int, ...]:
        """The shape of the voxel array."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_real_voxel_grid(
        cls, real_voxel_grid: Float[NDArrayLike, "dim dim dim"]
    ) -> Self:
        """Load an `AbstractVoxelVolume` from a 3D grid in
        real-space.
        """
        raise NotImplementedError


class AbstractVolumeIntegrator(eqx.Module, Generic[VolRep], strict=True):
    """Base class for a method of integrating a volume onto
    the exit plane.
    """

    outputs_ewald_sphere: eqx.AbstractClassVar[bool]

    @abc.abstractmethod
    def integrate(
        self,
        volume_representation: VolRep,
        image_config: AbstractImageConfig,
        outputs_real_space: bool = False,
    ) -> ProjectionOrEwaldSphereArray:
        raise NotImplementedError


class AbstractVolumeRenderFn(eqx.Module, Generic[VolRep], strict=True):
    """Base class for rendering a volume onto voxels."""

    shape: eqx.AbstractVar[tuple[int, int, int]]
    voxel_size: eqx.AbstractVar[Float[Array, ""]]

    @abc.abstractmethod
    def __call__(
        self,
        volume_representation: VolRep,
        *,
        outputs_real_space: bool = True,
        outputs_rfft: bool = False,
        fftshifted: bool = False,
    ) -> Array:
        raise NotImplementedError



import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float

from ..._internal import error_if_not_fractional
from ...jax_util import FloatLike
from ...ndimage import AbstractFourierOperator
from .._image_config import AbstractImageConfig
from .transfer_function import AbstractCTF


class AbstractTransferTheory(eqx.Module, strict=True):
    """A transfer theory for the weak-phase approximation. This class
    propagates the fourier spectrum of the object from a plane directly below it to
    the plane of the detector. In other terms, it computes a noiseless cryo-EM
    image from a 2D projection.
    """

    ctf: eqx.AbstractVar[AbstractCTF]


class ContrastTransferTheory(AbstractTransferTheory, strict=True):
    """A transfer theory for the weak-phase approximation. This class
    propagates the fourier spectrum of the object from a plane directly below it to
    the plane of the detector. In other terms, it computes a noiseless cryo-EM
    image from a 2D projection.
    """

    ctf: AbstractCTF
    envelope: AbstractFourierOperator | None
    amplitude_contrast_ratio: Float[Array, ""]
    phase_shift: Float[Array, ""]

    def __init__(
        self,
        ctf: AbstractCTF,
        envelope: AbstractFourierOperator | None = None,
        amplitude_contrast_ratio: FloatLike = 0.1,
        phase_shift: FloatLike = 0.0,
    ):
        """**Arguments:**

        - `ctf`: The contrast transfer function model.
        - `envelope`: The envelope function of the optics model.
        - `amplitude_contrast_ratio`: The amplitude contrast ratio.
        - `phase_shift`: The additional phase shift.
        """

        self.ctf = ctf
        self.envelope = envelope
        self.amplitude_contrast_ratio = jnp.asarray(amplitude_contrast_ratio, dtype=float)
        self.phase_shift = jnp.asarray(phase_shift, dtype=float)

    def propagate_object(
        self,
        object_spectrum: (
            Complex[
                Array,
                "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}",
            ]
            | Complex[
                Array,
                "{image_config.padded_y_dim} {image_config.padded_x_dim}",
            ]
        ),
        image_config: AbstractImageConfig,
        *,
        defocus_offset: FloatLike | None = None,
        input_is_ewald_sphere: bool = False,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim//2+1}"]:
        """Apply the CTF directly to the phase shifts in the exit plane.

        **Arguments:**

        - `object_spectrum`:
            The fourier spectrum of the scatterer phase shifts in a plane directly
            below it.
        - `image_config`:
            The configuration of the resulting image.
        - `input_is_ewald_sphere`:
            If `False`, the `object_spectrum_in_exit_plane` is a projection
            approximation and is therefore the fourier transform of a real-valued
            array. If `True`, `object_spectrum_in_exit_plane` is extracted from
            the ewald sphere and is therefore the fourier transform of a complex-valued
            array.
        - `defocus_offset`:
            An optional defocus offset to apply to the CTF defocus at
            runtime.
        """
        amplitude_contrast_ratio = error_if_not_fractional(self.amplitude_contrast_ratio)
        frequency_grid = image_config.get_frequency_grid(padding=True, physical=True)
        if not input_is_ewald_sphere:
            # Compute the CTF, including additional phase shifts
            ctf_array = self.ctf(
                frequency_grid,
                voltage_in_kilovolts=image_config.voltage_in_kilovolts,
                phase_shift=self.phase_shift,
                amplitude_contrast_ratio=amplitude_contrast_ratio,
                outputs_exp=False,
                defocus_offset=defocus_offset,
            )
            # ... compute the contrast as the CTF multiplied by the exit plane
            # phase shifts
            contrast_spectrum = ctf_array * object_spectrum
        else:
            # Propagate to the exit plane when the phase spectrum is
            # the surface of the ewald sphere
            aberration_phase_shifts = self.ctf.compute_aberration_phase_shifts(
                frequency_grid,
                wavelength_in_angstroms=image_config.wavelength_in_angstroms,
                defocus_offset=defocus_offset,
            ) - jnp.deg2rad(self.phase_shift)
            contrast_spectrum = _compute_contrast_from_ewald_sphere(
                object_spectrum,
                aberration_phase_shifts,
                amplitude_contrast_ratio,
                image_config,
            )
        if self.envelope is not None:
            contrast_spectrum *= self.envelope(frequency_grid)

        return contrast_spectrum


class WaveTransferTheory(AbstractTransferTheory, strict=True):
    """An optics model that propagates the exit wave to the detector plane."""

    ctf: AbstractCTF

    def __init__(
        self,
        ctf: AbstractCTF,
    ):
        """**Arguments:**

        - `ctf`: The contrast transfer function model.
        """

        self.ctf = ctf

    def propagate_exit_wave(
        self,
        wavefunction_spectrum: Complex[
            Array,
            "{image_config.padded_y_dim} {image_config.padded_x_dim}",
        ],
        image_config: AbstractImageConfig,
        *,
        defocus_offset: FloatLike | None = None,
    ) -> Complex[Array, "{image_config.padded_y_dim} {image_config.padded_x_dim}"]:
        """Apply the wave transfer function to the wavefunction in the exit plane."""
        frequency_grid = image_config.get_frequency_grid(padding=True, full=True)
        # Compute the wave transfer function
        ctf_array = self.ctf(
            frequency_grid,
            voltage_in_kilovolts=image_config.voltage_in_kilovolts,
            outputs_exp=True,
            defocus_offset=defocus_offset,
        )
        # ... compute the contrast as the CTF multiplied by the exit plane
        # phase shifts
        wavefunction_spectrum = ctf_array * wavefunction_spectrum

        return wavefunction_spectrum


def _compute_contrast_from_ewald_sphere(
    object_spectrum,
    aberration_phase_shifts,
    amplitude_contrast_ratio,
    image_config,
):
    cos, sin = jnp.cos(aberration_phase_shifts), jnp.sin(aberration_phase_shifts)
    ac = amplitude_contrast_ratio
    # Compute the contrast, breaking the computation into positive and
    # negative frequencies
    y_dim, x_dim = image_config.padded_y_dim, image_config.padded_x_dim
    # ... first handle the grid of frequencies
    pos_object_yx = object_spectrum[1:, 1 : x_dim // 2 + x_dim % 2]
    neg_object_yx = jnp.flip(
        jnp.flip(object_spectrum[1:, x_dim // 2 + x_dim % 2 :], axis=-1),
        axis=0,
    )
    if x_dim % 2 == 0:
        pos_object_yx = jnp.concatenate(
            (pos_object_yx, neg_object_yx[:, -1, None].conj()), axis=-1
        )
    contrast_yx = _ewald_propagate_kernel(
        pos_object_yx,
        neg_object_yx,
        ac,
        sin[1:, 1:],
        cos[1:, 1:],
    )
    # ... next handle the line of frequencies at y = 0
    pos_object_0x = object_spectrum[0, 1 : x_dim // 2 + x_dim % 2]
    neg_object_0x = jnp.flip(object_spectrum[0, x_dim // 2 + x_dim % 2 :], axis=-1)
    if x_dim % 2 == 0:
        pos_object_0x = jnp.concatenate(
            (pos_object_0x, neg_object_0x[-1, None].conj()), axis=0
        )
    contrast_0x = _ewald_propagate_kernel(
        pos_object_0x,
        neg_object_0x,
        ac,
        sin[0, 1 : x_dim // 2 + 1],
        cos[0, 1 : x_dim // 2 + 1],
    )
    # ... then handle the line of frequencies at x = 0
    pos_object_y0 = object_spectrum[1 : y_dim // 2 + y_dim % 2, 0]
    neg_object_y0 = jnp.flip(object_spectrum[y_dim // 2 + y_dim % 2 :, 0], axis=-1)
    if y_dim % 2 == 0:
        pos_object_y0 = jnp.concatenate(
            (pos_object_y0, neg_object_y0[-1, None].conj()), axis=0
        )
    contrast_y0 = _ewald_propagate_kernel(
        pos_object_y0,
        neg_object_y0,
        ac,
        sin[1 : y_dim // 2 + 1, 0],
        cos[1 : y_dim // 2 + 1, 0],
    )
    # ... concatenate the zero mode to the line of frequencies at x = 0
    object_00 = object_spectrum[0, 0]
    contrast_00 = _ewald_propagate_kernel(
        object_00,
        object_00,
        ac,
        sin[0, 0],
        cos[0, 0],
    )
    contrast_y0 = jnp.concatenate(
        (
            contrast_00[None],
            (contrast_y0 if y_dim % 2 == 1 else contrast_y0[:-1]),
            jnp.flip(contrast_y0.conjugate()),
        ),
        axis=0,
    )
    # ... concatenate the results
    contrast_yx = jnp.concatenate((contrast_0x[None, :], contrast_yx), axis=0)
    contrast_spectrum = jnp.concatenate((contrast_y0[:, None], contrast_yx), axis=1)

    return contrast_spectrum


def _ewald_propagate_kernel(pos, neg, ac, sin, cos):
    w1, w2 = ac, jnp.sqrt(1 - ac**2)
    return 0.5 * (
        (w2 * (pos.real + neg.real) + w1 * (pos.imag + neg.imag)) * sin
        + (w2 * (pos.imag + neg.imag) - w1 * (pos.real + neg.real)) * cos
        + 1.0j
        * (
            (w2 * (pos.imag - neg.imag) + w1 * (neg.real - pos.real)) * sin
            + (w2 * (neg.real - pos.real) + w1 * (neg.imag - pos.imag)) * cos
        )
    )



import os

import jax
import mdtraj
import mmdf
import numpy as np
import pandas as pd
import pytest
from jaxtyping import install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    from cryojax.io import mmdf_to_atoms, read_atoms_from_pdb


@pytest.fixture
def pdb_multiple_structures_path():
    return os.path.join(os.path.dirname(__file__), "data", "1uao_assembly.pdb")


def test_read_structure(sample_pdb_path):
    atom_positions, atomic_numbers, atom_properties = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="protein and not element H",
        loads_properties=True,
    )

    assert atom_positions.ndim == 2
    assert jax.tree.reduce(
        lambda x, y: x and y,
        jax.tree.map(lambda z: z.shape == atomic_numbers.shape, atom_properties),
    )
    assert atom_positions.shape[0] == atomic_numbers.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 77


def test_read_structure_no_properties(sample_pdb_path):
    atom_positions, atomic_numbers = read_atoms_from_pdb(
        sample_pdb_path,
        center=True,
        selection_string="protein and not element H",
        loads_properties=False,
    )

    assert atom_positions.ndim == 2
    assert atom_positions.shape[0] == atomic_numbers.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 77


def test_read_pdb_multiple_structures_stack(pdb_multiple_structures_path):
    atom_positions, atomic_numbers, atom_properties = read_atoms_from_pdb(
        pdb_multiple_structures_path,
        center=True,
        loads_properties=True,
        selection_string="all",
        model_index=None,
        stack_models=True,
    )
    assert atom_positions.ndim == 3
    assert atom_positions.shape[0] == 10
    assert atom_positions.shape[2] == 3
    assert atom_positions.shape[1] == 138
    assert jax.tree.reduce(
        lambda x, y: x and y,
        jax.tree.map(lambda z: z.shape == atom_positions.shape[0:2], atom_properties),
    )
    assert atomic_numbers.shape == (138,)


def test_read_pdb_multiple_structures_flat(pdb_multiple_structures_path):
    atom_positions, atomic_numbers, atom_properties = read_atoms_from_pdb(
        pdb_multiple_structures_path,
        center=True,
        loads_properties=True,
        selection_string="all",
        model_index=None,
        stack_models=False,
    )
    assert atom_positions.ndim == 2
    assert atomic_numbers.shape == (10 * 138,)
    assert atom_positions.shape == (10 * 138, 3)
    assert jax.tree.reduce(
        lambda x, y: x and y,
        jax.tree.map(lambda z: z.shape == atomic_numbers.shape, atom_properties),
    )


def test_bad_read_pdb_multiple_structures(pdb_multiple_structures_path):
    df = mmdf.read(pdb_multiple_structures_path)
    last_row = df.iloc[[-1]]
    df = pd.concat([df, last_row], ignore_index=True)
    with pytest.raises(ValueError):
        _ = mmdf_to_atoms(
            df,
            center=True,
            loads_properties=True,
            selection_string="all",
            model_index=None,
            stack_models=True,
        )


def test_read_pdb_at_structure(pdb_multiple_structures_path):
    atom_positions, atomic_numbers = read_atoms_from_pdb(
        pdb_multiple_structures_path,
        center=True,
        loads_b_factors=False,
        selection_string="all",
        model_index=1,
    )

    assert atom_positions.ndim == 2
    assert atom_positions.shape[0] == atomic_numbers.shape[0]
    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 138


def test_read_cif(sample_cif_path):
    atom_positions, atomic_numbers, atom_properties = read_atoms_from_pdb(
        sample_cif_path,
        center=True,
        selection_string="all",
        model_index=None,
        loads_properties=True,
    )

    assert atom_positions.ndim == 2
    assert jax.tree.reduce(
        lambda x, y: x and y,
        jax.tree.map(lambda z: z.shape == atomic_numbers.shape, atom_properties),
    )
    assert atom_positions.shape[0] == atomic_numbers.shape[0]

    assert atom_positions.shape[1] == 3
    assert atom_positions.shape[0] == 3222


def test_center_waterbox(sample_waterbox_pdb):
    atom_positions, _ = read_atoms_from_pdb(
        sample_waterbox_pdb,
        center=True,
        selection_string="all",
    )

    assert not np.isnan(atom_positions).any(), "Centering resulted in positions with NaNs"


@pytest.mark.parametrize("pdbfile", ["sample_pdb_path", "sample_waterbox_pdb"])
@pytest.mark.parametrize("center", [True, False])
def test_consistency_with_mdtraj(pdbfile, center, request):
    pdbfile = request.getfixturevalue(pdbfile)
    atom_positions, _ = read_atoms_from_pdb(
        pdbfile,
        center=center,
        selection_string="all",
    )

    import warnings

    warnings.filterwarnings("ignore")

    traj = mdtraj.load(pdbfile)
    if center:
        traj.center_coordinates(mass_weighted=True)

    mdtraj_positions = traj.xyz[0] * 10.0  # convert to Angstrom

    distance_matrix = np.linalg.norm(
        atom_positions[:, None, :] - mdtraj_positions[None, :, :], axis=-1
    )
    np.testing.assert_allclose(distance_matrix.min(axis=1), 0.0, atol=1e-4, rtol=1e-4)
    unique_min_indices = np.unique(distance_matrix.argmin(axis=1))
    assert unique_min_indices.shape[0] == atom_positions.shape[0], (
        "Atom position loading is not consistent with MDTraj up to permutation"
    )



"""Environmental variables for configuring cryoJAX behavior."""

import os


def _get_bool_var(varname: str, default: bool) -> bool:
    """Read an environmental variable and interpret it as a boolean.

    This function is modified from JAX:
    https://github.com/jax-ml/jax/blob/main/jax/_src/config.py.
    """
    val = os.getenv(varname, str(default))
    val = val.lower()
    if val in ("true", "1"):
        return True
    elif val in ("false", "0"):
        return False
    else:
        raise ValueError(
            f"Unrecognized value for `{varname}`. To enable `{varname}`, "
            "set to `true` or `1`. To disable, set to `false` or `0`."
        )


# Not public API
CRYOJAX_ENABLE_CHECKS: bool = _get_bool_var("CRYOJAX_ENABLE_CHECKS", False)



from ._lie_group import (
    SO2 as SO2,
    SO3 as SO3,
    AbstractMatrixLieGroup as AbstractMatrixLieGroup,
)
from ._rotation import AbstractRotation as AbstractRotation
from ._utils import (
    convert_quaternion_to_euler_angles as convert_quaternion_to_euler_angles,
)



import cryojax.simulator as cxs
import jax.numpy as jnp
import numpy as np
import pytest
from cryojax.rotations import SO3


def test_default_pose_arguments():
    euler = cxs.EulerAnglePose()
    quat = cxs.QuaternionPose()
    axis_angle = cxs.AxisAnglePose()
    np.testing.assert_allclose(euler.rotation.as_matrix(), quat.rotation.as_matrix())
    np.testing.assert_allclose(
        euler.rotation.as_matrix(), axis_angle.rotation.as_matrix()
    )


def test_translation_agreement():
    rotation = SO3(jnp.asarray((1.0, 0.0, 0.0, 0.0)))
    offset = jnp.asarray((0.0, -1.4))
    quat = cxs.QuaternionPose.from_rotation_and_translation(rotation, offset)
    axis_angle = cxs.AxisAnglePose.from_rotation_and_translation(rotation, offset)
    np.testing.assert_allclose(quat.rotation.as_matrix(), axis_angle.rotation.as_matrix())
    np.testing.assert_allclose(quat.offset_in_angstroms, axis_angle.offset_in_angstroms)


def test_pose_conversion():
    wxyz = jnp.asarray((1.0, 2.0, 3.0, 0.5))
    rotation = SO3(wxyz).normalize()
    quat = cxs.QuaternionPose.from_rotation(rotation)
    euler = cxs.EulerAnglePose.from_rotation(rotation)
    axis_angle = cxs.AxisAnglePose.from_rotation(rotation)
    np.testing.assert_allclose(quat.rotation.as_matrix(), euler.rotation.as_matrix())
    np.testing.assert_allclose(quat.rotation.as_matrix(), axis_angle.rotation.as_matrix())


def test_invert_pose():
    wxyz = jnp.asarray((1.0, 2.0, 3.0, 0.5))
    offset = jnp.asarray((-1.0, 1.0))
    rotation = SO3(wxyz).normalize()
    quat = cxs.QuaternionPose.from_rotation_and_translation(rotation, offset)
    quat_inverse = quat.to_inverse_rotation()
    np.testing.assert_allclose(quat.wxyz, quat_inverse.wxyz.at[1:].mul(-1))
    np.testing.assert_allclose(quat.offset_in_angstroms, quat_inverse.offset_in_angstroms)


def test_axis_angle_euler_agreement():
    angle = 2.0
    angle_in_radians = jnp.deg2rad(angle)
    rotation_x = SO3.from_x_radians(angle_in_radians)
    rotation_y = SO3.from_y_radians(angle_in_radians)
    rotation_z = SO3.from_z_radians(angle_in_radians)
    aa_x = cxs.AxisAnglePose(euler_vector=(angle, 0.0, 0.0))
    aa_y = cxs.AxisAnglePose(euler_vector=(0.0, angle, 0.0))
    aa_z = cxs.AxisAnglePose(euler_vector=(0.0, 0.0, angle))
    np.testing.assert_allclose(rotation_x.as_matrix(), aa_x.rotation.as_matrix())
    np.testing.assert_allclose(rotation_y.as_matrix(), aa_y.rotation.as_matrix())
    np.testing.assert_allclose(rotation_z.as_matrix(), aa_z.rotation.as_matrix())


@pytest.mark.parametrize(
    "phi, theta, psi",
    [
        (2.0, 15.0, -40.0),
        (10.0, 90.0, 170.0),
        (-10.0, 120.0, 140.0),
        (-120.0, 40.0, -80.0),
    ],
)
def test_euler_angle_conversion(phi, theta, psi):
    pose = cxs.EulerAnglePose(phi_angle=phi, theta_angle=theta, psi_angle=psi)
    converted_pose = cxs.EulerAnglePose.from_rotation(pose.rotation)
    np.testing.assert_allclose(
        np.asarray((phi, theta, psi)),
        np.asarray(
            (
                converted_pose.phi_angle,
                converted_pose.theta_angle,
                converted_pose.psi_angle,
            )
        ),
    )


@pytest.mark.parametrize(
    "wxyz",
    [
        (1.0, 0.0, 0.0, 0.0),
        (0.2, 0.4, 0.3, -0.1),
        (-0.1, -0.25, 0.6, -0.2),
    ],
)
def test_inverse(wxyz):
    wxyz = jnp.asarray(wxyz)
    wxyz = wxyz / jnp.linalg.norm(wxyz)
    euler = cxs.EulerAnglePose.from_rotation(SO3(wxyz))
    quat = cxs.QuaternionPose.from_rotation(SO3(wxyz))
    aa = cxs.AxisAnglePose.from_rotation(SO3(wxyz))
    wxyz_inverse = lambda pose: pose.to_inverse_rotation().rotation.wxyz
    check_quat_equal = lambda q1, q2: np.allclose(q1, q2) or np.allclose(q1, -q2)
    assert check_quat_equal(wxyz_inverse(euler), wxyz_inverse(quat))
    assert check_quat_equal(wxyz_inverse(quat), wxyz_inverse(aa))



from collections.abc import Callable, Iterable
from typing import Any, TypeVar

import equinox as eqx
import jax
import lineax as lx
from jaxtyping import Array, PyTree

from ._filter_specs import make_filter_spec


Args = TypeVar("Args")


def make_linear_operator(
    fn: Callable[[Args], Array],
    args: Args,
    where_vector: Callable[[Args], Any],
    *,
    tags: object | Iterable[object] = (),
) -> tuple[lx.FunctionLinearOperator, Args]:
    """Instantiate a [`lineax.FunctionLinearOperator`](https://docs.kidger.site/lineax/api/operators/#lineax.FunctionLinearOperator)
    to from a function that takes an arbitrary pytree as input.

    This is useful for converting from the cryoJAX abstraction for image simulation
    to a [`lineax`](https://docs.kidger.site/lineax/)'s matrix-vector multiplication
    abstraction. It is easy to get backprojection operators using `lineax`, which
    calls [`jax.linear_transpose`](https://docs.jax.dev/en/latest/_autosummary/jax.linear_transpose.html)
    under the hood.

    !!! example "Backprojection with `lineax`"

        ```python
        import cryojax.simulator as cxs
        import cryojax.jax_util as jxu
        import lineax as lx

        # Instantiate a linear operator
        volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(...)
        image_model = cxs.make_image_model(volume, ...)
        where_vector = lambda x: x.volume.fourier_voxel_grid
        operator, vector = jxu.make_linear_operator(
            fn=lambda x: x.simulate(),
            args=image_model,
            where_vector=where_vector,
        )
        # Simulate an image
        image = operator.mv(vector)
        # Compute backprojection
        adjoint = lx.conj(operator.T)
        backprojection = where_vector(adjoint.mv(image))
    ```

    !!! warning

        This function promises that `fn` can be expressed as a
        linear operator with respect to the input arguments at `where_vector`.
        CryoJAX does not explicitly check if this is the case, so JAX will
        throw errors downstream.

    **Arguments:**

    - `fn`:
        A function with signature `out = fn(args)`
    - `args`:
        Input arguments to `fn`
    - `where_vector`:
        A pointer to where the arguments for the volume
        input space are in `args`.
    - `tags`:
        See `lineax.FunctionLinearOperator` for documentation.

    **Returns:**

    A tuple with first element `lineax.FunctionLinearOperator` and second element
    a pytree with the same structure as `pytree`, partitioned to only include the
    arguments at `where_vector`.
    """  # noqa: E501
    # Extract arguments for the volume at `where_vector`
    filter_spec = make_filter_spec(args, where_vector)
    vector_args, other_args = eqx.partition(args, filter_spec)
    vector, static_args = eqx.partition(vector_args, eqx.is_array)
    other_args = eqx.combine(other_args, static_args)
    # Instantiate the `lineax.FunctionLinearOperator`
    fn_wrapper = _FnWrapper(fn, other_args)
    input_structure = jax.tree.map(
        lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), vector
    )
    linear_operator = lx.FunctionLinearOperator(
        fn=fn_wrapper, input_structure=input_structure, tags=tags
    )
    return linear_operator, vector


class _FnWrapper(eqx.Module):
    fn: Callable[[PyTree], Array]
    args: PyTree

    def __call__(self, vector: PyTree) -> Array:
        args = eqx.combine(vector, self.args)
        return self.fn(args)



from .custom_types import (
    PyTreeGrid as PyTreeGrid,
    PyTreeGridIndex as PyTreeGridIndex,
    PyTreeGridPoint as PyTreeGridPoint,
)
from .pytree_manipulation import (
    tree_grid_shape as tree_grid_shape,
    tree_grid_take as tree_grid_take,
    tree_grid_unravel_index as tree_grid_unravel_index,
)
from .search_loop import run_grid_search as run_grid_search
from .search_method import (
    AbstractGridSearchMethod as AbstractGridSearchMethod,
    MinimumSearchMethod as MinimumSearchMethod,
)



"""cryoJAX compatibility with [RELION](https://relion.readthedocs.io/en/release-5.0/)."""

import abc
import pathlib
import re
import warnings
from collections.abc import Callable
from copy import deepcopy
from typing import Any, Literal, TypedDict, cast
from typing_extensions import Self, override

import equinox as eqx
import jax
import jax.numpy as jnp
import mrcfile
import numpy as np
import pandas as pd
from jaxtyping import Array, Float, Int

from ...io import read_starfile, write_image_stack_to_mrc, write_starfile
from ...jax_util import NDArrayLike
from ...ndimage import FourierConstant, FourierGaussian
from ...simulator import (
    AstigmaticCTF,
    BasicImageConfig,
    ContrastTransferTheory,
    EulerAnglePose,
)
from .base_particle_dataset import AbstractParticleDataset, AbstractParticleParameterFile


# RELION column entries
RELION_CTF_OPTICS_ENTRIES = [
    ("rlnSphericalAberration", "Float64"),
    ("rlnAmplitudeContrast", "Float64"),
]
RELION_INSTRUMENT_OPTICS_ENTRIES = [
    ("rlnImageSize", "Int64"),
    ("rlnVoltage", "Float64"),
    ("rlnImagePixelSize", "Float64"),
]
RELION_CTF_PARTICLE_ENTRIES = [
    ("rlnDefocusU", "Float64"),
    ("rlnDefocusV", "Float64"),
    ("rlnDefocusAngle", "Float64"),
    ("rlnPhaseShift", "Float64"),
]
RELION_POSE_PARTICLE_ENTRIES = [
    ("rlnOriginXAngst", "Float64"),
    ("rlnOriginYAngst", "Float64"),
    ("rlnAngleRot", "Float64"),
    ("rlnAngleTilt", "Float64"),
    ("rlnAnglePsi", "Float64"),
]
# Default entries for writing
RELION_DEFAULT_OPTICS_ENTRIES = [
    *RELION_INSTRUMENT_OPTICS_ENTRIES,
    *RELION_CTF_OPTICS_ENTRIES,
    ("rlnOpticsGroup", "Int64"),
]
RELION_DEFAULT_PARTICLE_ENTRIES = [
    *RELION_CTF_PARTICLE_ENTRIES,
    *RELION_POSE_PARTICLE_ENTRIES,
    ("rlnOpticsGroup", "Int64"),
]
# Required entries for loading
RELION_REQUIRED_OPTICS_ENTRIES = RELION_DEFAULT_OPTICS_ENTRIES
RELION_REQUIRED_PARTICLE_ENTRIES = [
    *RELION_CTF_PARTICLE_ENTRIES,
    ("rlnOpticsGroup", "Int64"),
]
RELION_SUPPORTED_PARTICLE_ENTRIES = [
    *RELION_REQUIRED_PARTICLE_ENTRIES,
    *RELION_POSE_PARTICLE_ENTRIES,
    ("rlnCtfBfactor", "Float64"),
    ("rlnCtfScalefactor", "Float64"),
]


class _ParticleParameterInfo(TypedDict):
    """Parameters for a particle stack from RELION."""

    image_config: BasicImageConfig
    pose: EulerAnglePose
    transfer_theory: ContrastTransferTheory

    metadata: pd.DataFrame | None


class _ParticleStackInfo(TypedDict):
    """Particle stack info from RELION."""

    parameters: _ParticleParameterInfo | None
    images: Float[np.ndarray, "... y_dim x_dim"]


_ParticleParameterLike = dict[str, Any] | _ParticleParameterInfo
_ParticleStackLike = dict[str, Any] | _ParticleStackInfo


class _StarfileData(TypedDict):
    optics: pd.DataFrame
    particles: pd.DataFrame


class _MrcfileSettings(TypedDict):
    prefix: str
    output_folder: str | pathlib.Path
    n_characters: int
    delimiter: str
    overwrite: bool
    compression: str | None


class AbstractParticleStarFile(
    AbstractParticleParameterFile[_ParticleParameterInfo, _ParticleParameterLike]
):
    @property
    @override
    def path_to_output(self) -> pathlib.Path:
        return self.path_to_starfile

    @path_to_output.setter
    @override
    def path_to_output(self, value: str | pathlib.Path):
        self.path_to_starfile = value

    @property
    @abc.abstractmethod
    def path_to_starfile(self) -> pathlib.Path:
        raise NotImplementedError

    @path_to_starfile.setter
    @abc.abstractmethod
    def path_to_starfile(self, value: str | pathlib.Path):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def starfile_data(self) -> _StarfileData:
        raise NotImplementedError

    @starfile_data.setter
    @abc.abstractmethod
    def starfile_data(self, value: dict[str, pd.DataFrame]):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def loads_metadata(self) -> bool:
        raise NotImplementedError

    @loads_metadata.setter
    @abc.abstractmethod
    def loads_metadata(self, value: bool):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def loads_envelope(self) -> bool:
        raise NotImplementedError

    @loads_envelope.setter
    @abc.abstractmethod
    def loads_envelope(self, value: bool):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def broadcasts_image_config(self) -> bool:
        raise NotImplementedError

    @broadcasts_image_config.setter
    @abc.abstractmethod
    def broadcasts_image_config(self, value: bool):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def updates_optics_group(self) -> bool:
        raise NotImplementedError

    @updates_optics_group.setter
    @abc.abstractmethod
    def updates_optics_group(self, value: bool):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def rotation_convention(self) -> Literal["object", "frame"]:
        raise NotImplementedError

    @rotation_convention.setter
    @abc.abstractmethod
    def rotation_convention(self, value: Literal["object", "frame"]):
        raise NotImplementedError

    def copy(self) -> Self:
        return deepcopy(self)


class RelionParticleParameterFile(AbstractParticleStarFile):
    """A dataset that wraps a RELION particle stack in
    [STAR](https://relion.readthedocs.io/en/latest/Reference/Conventions.html)
    format.
    """

    def __init__(
        self,
        path_to_starfile: str | pathlib.Path,
        mode: Literal["r", "w"] = "r",
        exists_ok: bool = False,
        selection_filter: dict[str, Callable] | None = None,
        *,
        loads_metadata: bool = False,
        broadcasts_image_config: bool = True,
        loads_envelope: bool = False,
        updates_optics_group: bool = False,
        rotation_convention: Literal["frame", "object"] = "object",
        pad_options: dict = {},
    ):
        """**Arguments:**

        - `path_to_starfile`:
            The path to the RELION STAR file. If the path does not exist
            and `mode = 'w'`, an empty dataset will be created.
        - `path_to_relion_project`: The path to the RELION project directory.
        - `mode`:
            - If `mode = 'w'`, the dataset is prepared to write new
            *parameters*. This is done by storing an empty dataset in
            `RelionParticleParameterFile.starfile_data`. If a STAR file
            already exists at `path_to_starfile`, set `exists_ok = True`.
            - If `mode = 'r'`, the STAR file at `path_to_starfile` is read
            into `RelionParticleParameterFile.starfile_data`.
        - `exists_ok`:
            If the `path_to_starfile` already exists, if `True` and `mode = 'w'`
            nonetheless stores an empty `RelionParticleParameterFile.starfile_data`.
        - `selection_filter`:
            A dictionary used to include only particular dataset elements.
            The keys of this dictionary should be any data entry in the STAR
            file, while the values should be a function that takes in a
            column and returns a boolean mask for the column. For example,
            filter by class using
            `selection_filter["rlnClassNumber"] = lambda x: x == 0`.
        - `loads_metadata`:
            If `True`, the resulting dict loads
            the raw metadata from the STAR file that is not otherwise included
            into a `pandas.DataFrame`.
            If this is set to `True`, note that dictionaries cannot pass through
            JIT boundaries without removing the metadata.
        - `broadcasts_image_config`:
            If `True`, image config parameters are broadcasted with leading dimension
            as the number of particles.
        - `loads_envelope`:
            If `True`, read in the parameters of the CTF envelope function, i.e.
            "rlnCtfScalefactor" and "rlnCtfBfactor".
        - `updates_optics_group`:
            If `True`, when re-writing STAR file entries via
            `dataset[idx] = parameters` syntax, creates a new optics group entry.
        - `rotation_convention`:
            If `'object'`, the loader loads/writes poses in the convention that
            the rotation is of the *object*. If `'frame'`, the rotation is of
            the frame (i.e. the rotation is inverted).
            The pose passed to [`cryojax.simulator.make_image_model`][]
            is always of the object, but advanced considerations may require
            setting `rotation_convention = 'frame'` (or manually calling
            `pose.to_inverse_rotation()`) to correctly match RELION and
            cryoJAX conventions.
        - `pad_options`:
            Padding options for image simulation, passed to the `BasicImageConfig`.
            See `BasicImageConfig` for documentation.
        """
        # Private attributes
        self._pad_options = pad_options
        self._mode = _validate_mode(mode)
        # The STAR file data
        self._path_to_starfile = pathlib.Path(path_to_starfile)
        self._starfile_data = _load_starfile_data(
            self._path_to_starfile, selection_filter, mode, exists_ok
        )
        # Properties for loading
        self._loads_metadata = loads_metadata
        self._broadcasts_image_config = broadcasts_image_config
        self._loads_envelope = loads_envelope
        # Properties for writing
        self._updates_optics_group = updates_optics_group
        # Shared
        self._rotation_convention = _validate_rotation_convention(rotation_convention)

    @override
    def __getitem__(
        self, index: int | slice | Int[np.ndarray, ""] | Int[np.ndarray, " _"]
    ) -> _ParticleParameterInfo:
        # Validate index
        n_rows = self.starfile_data["particles"].shape[0]
        _validate_dataset_index(type(self), index, n_rows)
        # ... read particle data at the requested indices
        particle_data = self.starfile_data["particles"]
        particle_data_at_index = particle_data.iloc[index]
        # ... read optics group given the particle data
        optics_data = self.starfile_data["optics"]
        optics_group = _get_optics_group_from_particle_data(
            particle_data_at_index, optics_data
        )
        # Load the image stack and STAR file parameters
        image_config, transfer_theory, pose = _make_pytrees_from_starfile(
            particle_data_at_index,
            optics_group,
            self.broadcasts_image_config,
            self.loads_envelope,
            self._pad_options,
            self._rotation_convention,
        )
        if self.loads_metadata:
            # ... convert to dataframe for serialization
            if isinstance(particle_data_at_index, pd.Series):
                particle_data_at_index = particle_data_at_index.to_frame().T
            # ... no overlapping keys with loaded pytrees
            redundant_entry_labels, _ = list(zip(*RELION_SUPPORTED_PARTICLE_ENTRIES))
            columns = particle_data_at_index.columns
            remove_columns = [
                column for column in columns if column in redundant_entry_labels
            ]
            metadata = particle_data_at_index.drop(remove_columns, axis="columns")
        else:
            metadata = None

        return _ParticleParameterInfo(
            image_config=image_config,
            pose=pose,
            transfer_theory=transfer_theory,
            metadata=metadata,
        )

    @override
    def __len__(self) -> int:
        return len(self.starfile_data["particles"])

    @override
    def __setitem__(
        self,
        index: int | slice | Int[np.ndarray, ""] | Int[np.ndarray, " _"],
        value: _ParticleParameterLike,
    ):
        # Make sure index is valid
        n_rows = self.starfile_data["particles"].shape[0]
        _validate_dataset_index(type(self), index, n_rows)
        # ... also, the parameters too
        _validate_parameters(value, force_keys=False)
        # Invert the pose if desired
        if "pose" in value:
            if self.rotation_convention == "frame":
                value["pose"] = _invert_rotation(value["pose"])
        # Grab the current and new optics and particle data
        if self.updates_optics_group:
            optics_group_index = _make_optics_group_index(self.starfile_data["optics"])
            particle_data_for_update = _parameters_to_particle_data(
                value, optics_group_index
            )
            optics_data_to_append = _parameters_to_optics_data(value, optics_group_index)
            optics_data = pd.concat(
                [self.starfile_data["optics"], optics_data_to_append], ignore_index=True
            )
        else:
            particle_data_for_update = _parameters_to_particle_data(value)
            optics_data = self.starfile_data["optics"]
        particle_data = self.starfile_data["particles"]
        # Set new empty columns in the particle data, if the update data includes this
        new_columns = list(
            set(particle_data_for_update.columns) - set(particle_data.columns)
        )
        for column in new_columns:
            dtype = pd.api.types.pandas_dtype(particle_data_for_update[column].dtype)
            particle_data[column] = pd.Series(dtype=dtype)
        # Finally, set the updated data
        if isinstance(index, (int, np.ndarray)):
            index = np.atleast_1d(index)
        particle_data.loc[
            particle_data.index[index], particle_data_for_update.columns
        ] = particle_data_for_update.values
        self._starfile_data = _StarfileData(optics=optics_data, particles=particle_data)

    @override
    def append(self, value: _ParticleParameterLike):
        """Add an entry or entries to the dataset.

        **Arguments:**

        - `value`:
            A dictionary of parameters to add to the dataset.
        """
        # Make sure parameters are valid
        _validate_parameters(value, force_keys=True)
        # Invert the pose if desired
        if "pose" in value:
            if self.rotation_convention == "frame":
                value["pose"] = _invert_rotation(value["pose"])
        # Make new optics group
        optics_group_index = _make_optics_group_index(self.starfile_data["optics"])
        optics_data, optics_data_to_append = (
            self.starfile_data["optics"],
            _parameters_to_optics_data(value, optics_group_index),
        )
        # Make new particle entries
        particle_data, particle_data_to_append = (
            self.starfile_data["particles"],
            _parameters_to_particle_data(value, optics_group_index),
        )
        # Concatenate and set new entries
        optics_data = (
            pd.concat([optics_data, optics_data_to_append], ignore_index=True)
            if len(optics_data) > 0
            else optics_data_to_append
        )
        particle_data = (
            pd.concat([particle_data, particle_data_to_append], ignore_index=True)
            if len(particle_data) > 0
            else particle_data_to_append
        )
        self._starfile_data = _StarfileData(optics=optics_data, particles=particle_data)

    @override
    def save(
        self,
        *,
        overwrite: bool = False,
        **kwargs: Any,
    ):
        path_to_starfile = self.path_to_starfile
        path_exists = path_to_starfile.exists()
        if path_exists and not overwrite:
            raise FileExistsError(
                f"Tried saving STAR file, but file {str(path_to_starfile)} "
                "already exists. To overwrite existing STAR file, set "
                f"`{type(self).__name__}.overwrite = True`."
            )
        else:
            if not path_to_starfile.parent.exists():
                path_to_starfile.parent.mkdir(parents=True)
            write_starfile(self.starfile_data, path_to_starfile, **kwargs)

    @property
    @override
    def path_to_starfile(self) -> pathlib.Path:
        return self._path_to_starfile

    @path_to_starfile.setter
    @override
    def path_to_starfile(self, value: str | pathlib.Path):
        self._path_to_starfile = pathlib.Path(value)

    @property
    def mode(self) -> Literal["r", "w"]:
        return self._mode  # type: ignore

    @property
    @override
    def starfile_data(self) -> _StarfileData:
        return self._starfile_data

    @starfile_data.setter
    @override
    def starfile_data(self, value: dict[str, pd.DataFrame]):
        if "particles" in value and "optics" in value:
            particle_data, optics_data = value["particles"], value["optics"]
            if isinstance(particle_data, pd.DataFrame) and isinstance(
                optics_data, pd.DataFrame
            ):
                self._starfile_data = _StarfileData(
                    optics=optics_data, particles=particle_data
                )
            else:
                raise ValueError(
                    "STAR file data must be a dictionary "
                    "of pandas DataFrames, with keys equal to "
                    "'particles' and 'optics'. Found that the "
                    f"particle data was type `{type(particle_data).__name__}`"
                    "and the optics data was type "
                    f"`{type(optics_data).__name__}`."
                )
        else:
            raise ValueError(
                "STAR file data must be a dictionary "
                "of pandas DataFrames, with keys equal to "
                "'particles' and 'optics'. Tried setting "
                f"with a dictionary with keys {list(value.keys())}."
            )

    @property
    @override
    def loads_metadata(self) -> bool:
        return self._loads_metadata

    @loads_metadata.setter
    @override
    def loads_metadata(self, value: bool):
        self._loads_metadata = value

    @property
    @override
    def loads_envelope(self) -> bool:
        return self._loads_envelope

    @loads_envelope.setter
    @override
    def loads_envelope(self, value: bool):
        self._loads_envelope = value

    @property
    @override
    def broadcasts_image_config(self) -> bool:
        return self._broadcasts_image_config

    @broadcasts_image_config.setter
    @override
    def broadcasts_image_config(self, value: bool):
        self._broadcasts_image_config = value

    @property
    @override
    def updates_optics_group(self) -> bool:
        return self._updates_optics_group

    @updates_optics_group.setter
    @override
    def updates_optics_group(self, value: bool):
        self._updates_optics_group = value

    @property
    def rotation_convention(self) -> Literal["object", "frame"]:
        return self._rotation_convention  # pyright: ignore[reportReturnType]

    @rotation_convention.setter
    def rotation_convention(self, value: Literal["object", "frame"]):
        self._rotation_convention = _validate_rotation_convention(value)


class RelionParticleDataset(
    AbstractParticleDataset[_ParticleStackInfo, _ParticleStackLike]
):
    """A dataset that wraps a RELION particle stack in
    [STAR](https://relion.readthedocs.io/en/latest/Reference/Conventions.html) format.
    """

    def __init__(
        self,
        parameter_file: AbstractParticleStarFile,
        path_to_relion_project: str | pathlib.Path,
        mode: Literal["r", "w"] = "r",
        mrcfile_settings: dict[str, Any] = {},
        *,
        loads_parameters: bool = True,
    ):
        """**Arguments:**

        - `path_to_relion_project`:
            In RELION STAR files, only a relative path is added to the
            'rlnImageName' column. This is relative to the path to the
            "project", which is given by this parameter.
        - `parameter_file`:
            The `RelionParticleParameterFile`.
        - `mode`:
            - If `mode = 'w'`, the dataset is prepared to write new
            *images*. This is done by removing 'rlnImageName' from
            `parameter_file.starfile_data`, if it exists at all.
            does not have a column 'rlnImageName' and image files
            are not yet written.
            - If `mode = 'r'`, images are read from the 'rlnImageName'
            stored in the `parameter_file.starfile_data`.
        - `mrcfile_settings`:
            A dictionary with the following keys:
            - 'prefix':
                A `str` which acts as the prefix to the filenames. If this
                is equal to `"f"`, then the filename for image stack 0 will
                be called "f-00000.mrcs", for `delimiter = '-'` and
                `n_characters = 5`.
                are of format "filenam"
            - 'output_folder':
                A `str` or `pathlib.Path` type where to write MRC files,
                relative to the `path_to_relion_project`.
            - 'n_characters':
                An `int` for the number of characters to write the filename
                number string. If this is equal to `5`, then the filename
                for image stack 0 will be called "f-00000.mrcs", for
                `delimiter = '-'` and `prefix = 'f'`.
             - 'delimiter':
                A `str` for the delimiter between the filename prefix
                and number string. If this is equal to `'-'`, then the
                filename for image stack 0 will be called "f-00000.mrcs",
                for `n_characters = 5` and `prefix = 'f'`.
            - 'overwrite':
                If `True`, overwrite existing MRC file path if it exists.
        - `loads_parameters`:
            If `True`, load parameters and images. Otherwise, load only images.
        """
        # Set properties. First, core properties of the dataset, starting
        # with the `mode``
        self._mode = _validate_mode(mode)
        # ... then, the `parameter_file`. If `mode = 'w'` but
        # the images already exist, we should make a copy in case
        # those images are being used elsewhere
        particle_data, optics_data = (
            parameter_file.starfile_data["particles"],
            parameter_file.starfile_data["optics"],
        )
        self._parameter_file = parameter_file
        # ... properties common to reading and writing images
        self._path_to_relion_project = pathlib.Path(path_to_relion_project)
        # ... properties for reading images
        self._loads_parameters = loads_parameters
        # ... properties for writing images
        self._mrcfile_settings = _dict_to_mrcfile_settings(mrcfile_settings)
        # Now, initialize for `mode = 'r'` vs `mode = 'w'`
        images_exist = "rlnImageName" in particle_data.columns
        project_exists = self.path_to_relion_project.exists()
        if mode == "w":
            # Write empty "rlnImageName" column (defaults to NaN values)
            particle_data["rlnImageName"] = pd.Series(dtype=str)
            self.parameter_file.starfile_data = dict(
                optics=optics_data, particles=particle_data
            )
            # Make the project directory, if it does not yet exist
            if not project_exists:
                self.path_to_relion_project.mkdir(parents=True, exist_ok=False)
        else:
            if not images_exist:
                raise OSError(
                    "Could not find column 'rlnImageName' in the STAR file. "
                    "When using `mode = 'r'`, the STAR file must have this "
                    "column. To write images in a STAR file, "
                    "set `mode = 'w'`."
                )
            if not project_exists:
                raise FileNotFoundError(
                    "`RelionParticleDataset` opened in "
                    "'mode = `r`', but the RELION project directory "
                    "`path_to_relion_project` does not exist. "
                    "To write images in a STAR file in a new RELION project, "
                    "set `mode = 'w'`."
                )

    @override
    def __getitem__(
        self, index: int | slice | Int[np.ndarray, ""] | Int[np.ndarray, " N"]
    ) -> _ParticleStackInfo:
        if self.loads_parameters:
            # Load images and parameters. First, read parameters
            # and metadata from the STAR file
            loads_metadata = self.parameter_file.loads_metadata
            self.parameter_file.loads_metadata = True
            # ... read parameters
            parameters = self.parameter_file[index]
            # ... validate the metadata
            particle_data_at_index = cast(pd.DataFrame, parameters["metadata"])
            _validate_rln_image_name_exists(particle_data_at_index, index)
            # ... reset boolean to original value
            self.parameter_file.loads_metadata = loads_metadata
            if not loads_metadata:
                parameters["metadata"] = None
            # ... grab shape
            shape = parameters["image_config"].shape
            # ... load stack of images
            images = _load_image_stack_from_mrc(
                shape, particle_data_at_index, self.path_to_relion_project
            )
            # ... make sure images and parameters have same leading dim
            if parameters["pose"].offset_x_in_angstroms.ndim == 0:
                images = np.squeeze(images)

            return _ParticleStackInfo(parameters=parameters, images=images)
        else:
            # Otherwise, do not read parameters to more efficiently read images. First,
            # validate the dataset index.
            n_rows = self.parameter_file.starfile_data["particles"].shape[0]
            _validate_dataset_index(type(self), index, n_rows)
            # ... read particle data at the requested indices
            particle_data = self.parameter_file.starfile_data["particles"]
            particle_data_at_index = particle_data.iloc[index]
            if isinstance(particle_data_at_index, pd.Series):
                particle_data_at_index = particle_data_at_index.to_frame().T
            _validate_rln_image_name_exists(particle_data_at_index, index)
            # ... grab shape by reading the optics group
            optics_data = self.parameter_file.starfile_data["optics"]
            optics_group = _get_optics_group_from_particle_data(
                particle_data_at_index, optics_data
            )
            image_size = int(optics_group["rlnImageSize"])
            shape = (image_size, image_size)
            # ... load stack of images
            images = _load_image_stack_from_mrc(
                shape, particle_data_at_index, self.path_to_relion_project
            )
            # ... make sure image leading dim matches with index query
            if isinstance(index, int) or (
                isinstance(index, np.ndarray) and index.size == 0
            ):
                images = np.squeeze(images)

            return _ParticleStackInfo(parameters=None, images=images)

    @override
    def __len__(self) -> int:
        return len(self.parameter_file)

    @override
    def __setitem__(
        self, index: int | slice | Int[np.ndarray, ""], value: _ParticleStackLike
    ):
        if isinstance(index, Int[np.ndarray, "_"]):  # type: ignore
            raise ValueError(
                "When setting `dataset[index] = ...`, "
                "it is not supported to pass `index` as a 1D numpy-array."
            )
        if not isinstance(value, dict):
            raise TypeError(
                "When setting `dataset[index] = foo`, "
                "foo must be a dictionary with key "
                "'images' and optionally key 'parameters'."
            )
        images, parameters = _unpack_particle_stack_dict(value)
        if parameters is not None:
            self.parameter_file[index] = parameters
        n_particles = len(self.parameter_file)
        index_array = np.atleast_1d(_index_to_array(index, n_particles))
        self.write_images(index_array, images, parameters=parameters)

    @override
    def append(self, value: _ParticleStackLike):
        """Add an entry or entries to the dataset.

        **Arguments:**

        - `value`:

        """
        if not isinstance(value, dict):
            raise TypeError(
                "When appending `dataset.append(foo)`, "
                "`foo` must be a dictionary with keys "
                "'images' and 'parameters'."
            )
        images, parameters = _unpack_particle_stack_dict(value)
        if parameters is None:
            raise ValueError(
                "When appending dictionary `foo` as `dataset.append(foo)`, "
                "`foo` must have key 'parameters'."
            )
        start = len(self.parameter_file.starfile_data["particles"])
        # Append parameters. This automatically sets the 'rlnImageName'
        # column to NaNs
        self.parameter_file.append(parameters)
        # Write images
        stop = len(self.parameter_file.starfile_data["particles"])
        index_array = np.arange(start, stop, dtype=int)
        self.write_images(index_array, images, parameters=parameters)

    @override
    def write_images(
        self,
        index_array: Int[np.ndarray, " _"],
        images: Float[NDArrayLike, "... _ _"],
        parameters: _ParticleParameterLike | None = None,
    ):
        # Get relevant metadata
        particle_data = self.parameter_file.starfile_data["particles"]
        optics_data = self.parameter_file.starfile_data["optics"]
        if parameters is None:
            optics_group = _get_optics_group_from_particle_data(
                particle_data.iloc[index_array], optics_data
            )
            pixel_size, dim = (
                float(optics_group["rlnImagePixelSize"]),
                int(optics_group["rlnImageSize"]),
            )
        else:
            pixel_size, dim = (
                float(np.atleast_1d(parameters["image_config"].pixel_size)[0]),
                parameters["image_config"].shape[0],
            )
        if not (images.ndim in [2, 3] and images.shape[-2:] == (dim, dim)):
            raise ValueError(
                "Image(s) must be of "
                "shape `(n_images, dim, dim)` or `(dim, dim)`. "
                f"Tried writing image(s) of "
                f"shape {images.shape}."
            )
        # Prepare to write images
        if images.ndim == 2:
            images = images[None, ...]
        n_images, _ = images.shape[0], images.shape[1]
        n_particles = len(self.parameter_file)

        # Convert index into 1D ascending numpy array
        n_indices = index_array.size
        if n_images != n_indices:
            raise ValueError(
                "Tried to set dataset elements with an inconsistent number "
                f"of images. Found that the number of images was {n_images}, "
                f"while the number of dataset indices was {n_indices}."
            )
        # Get absolute path to the filename, as well as the 'rlnImageName'
        # column
        path_to_filename, rln_image_names = _make_image_filename(
            index_array,
            particle_data,
            n_particles,
            self.mrcfile_settings,
            self.path_to_relion_project,
        )
        # Set the STAR file column
        particle_data.loc[particle_data.index[index_array], "rlnImageName"] = (
            rln_image_names
        )
        self.parameter_file.starfile_data = dict(
            particles=particle_data, optics=optics_data
        )
        # ... and write the images to disk
        try:
            write_image_stack_to_mrc(
                images.astype(jnp.float32),
                pixel_size,
                path_to_filename,
                overwrite=self.mrcfile_settings["overwrite"],
                compression=self.mrcfile_settings["compression"],
            )
        except Exception as err:
            raise OSError(
                "Error occurred when writing image stack to MRC "
                "file. Most likely, the filename the writer "
                f"chose ({str(path_to_filename)}) already "
                "exists. Try changing the "
                "`RelionParticleDataset.mrcfile_settings`. "
                f"The error message was:\n{err}"
            )

    @property
    @override
    def parameter_file(self) -> AbstractParticleStarFile:
        return self._parameter_file

    @property
    @override
    def mode(self) -> Literal["r", "w"]:
        return self._mode  # type: ignore

    @property
    def path_to_relion_project(self) -> pathlib.Path:
        return self._path_to_relion_project

    @property
    def mrcfile_settings(self) -> _MrcfileSettings:
        return self._mrcfile_settings

    @mrcfile_settings.setter
    def mrcfile_settings(self, value: dict[str, Any]):
        self._mrcfile_settings = _dict_to_mrcfile_settings(value)

    @property
    def loads_parameters(self) -> bool:
        return self._loads_parameters

    @loads_parameters.setter
    def loads_parameters(self, value: bool):
        self._loads_parameters = value


def _load_starfile_data(
    path_to_starfile: pathlib.Path,
    selection_filter: dict[str, Callable] | None,
    mode: Literal["r", "w"],
    exists_ok: bool,
) -> _StarfileData:
    if mode == "r":
        if path_to_starfile.exists():
            starfile_data = read_starfile(path_to_starfile)
            _validate_starfile_data(starfile_data)
            if selection_filter is not None:
                starfile_data = _select_particles(starfile_data, selection_filter)
        else:
            raise FileNotFoundError(
                f"Set `mode = '{mode}'`, but STAR file {str(path_to_starfile)} does not "
                "exist. To write a new STAR file, set `mode = 'w'`."
            )
    else:
        if path_to_starfile.exists() and not exists_ok:
            raise FileExistsError(
                f"Set `mode = 'w'`, but STAR file {str(path_to_starfile)} already "
                "exists. To read an existing STAR file, set `mode = 'r'` or "
                "to erase an existing STAR file, set `mode = 'w'` and "
                "`exists_ok=True`."
            )
        else:
            if selection_filter is None:
                starfile_data = dict(
                    optics=pd.DataFrame(
                        data={
                            column: pd.Series(dtype=dtype)
                            for column, dtype in RELION_DEFAULT_OPTICS_ENTRIES
                        }
                    ),
                    particles=pd.DataFrame(
                        data={
                            column: pd.Series(dtype=dtype)
                            for column, dtype in RELION_DEFAULT_PARTICLE_ENTRIES
                        }
                    ),
                )
            else:
                raise ValueError(
                    "Initialized a `RelionParticleParameterFile` in `mode = 'w'` "
                    "but also passed a `selection_filter`. Selection is only used "
                    "in `mode = 'r'`."
                )

    return _StarfileData(
        optics=starfile_data["optics"], particles=starfile_data["particles"]
    )


def _validate_mode(mode: str) -> Literal["r", "w"]:
    if mode not in ["r", "w"]:
        raise ValueError(
            f"Passed unsupported `mode = {mode}`. Supported modes are 'r' and 'w'."
        )
    return mode  # type: ignore


def _validate_rotation_convention(mode: str) -> Literal["frame", "object"]:
    if mode not in ["frame", "object"]:
        raise ValueError(
            f"Passed unsupported `rotation_convention = {mode}`. "
            "Supported modes are 'object' and 'frame'."
        )
    return mode  # type: ignore


def _select_particles(
    starfile_data: dict[str, pd.DataFrame], selection_filter: dict[str, Callable]
) -> dict[str, pd.DataFrame]:
    particle_data = starfile_data["particles"]
    boolean_mask = pd.Series(True, index=particle_data.index)
    for key in selection_filter:
        if key in particle_data.columns:
            fn = selection_filter[key]
            column = particle_data[key]
            base_error_message = (
                f"Error filtering key '{key}' in the `selection_filter`. "
                f"To filter the STAR file entries, `selection_filter['{key}']`"
                "must be a function that takes in an array and returns a "
                "boolean mask."
            )
            if isinstance(selection_filter[key], Callable):
                try:
                    mask_at_column = fn(column)
                except Exception as err:
                    raise ValueError(
                        f"{base_error_message} "
                        "When calling the function, caught an error:\n"
                        f"{err}"
                    )
                if not pd.api.types.is_bool_dtype(mask_at_column):
                    raise ValueError(
                        f"{base_error_message} "
                        "Found that the function did not return "
                        "a boolean dtype."
                    )
            else:
                raise ValueError(base_error_message)
            # Update mask
            boolean_mask = mask_at_column & boolean_mask
        else:
            raise ValueError(
                f"Included key '{key}' in the `selection_filter`, "
                "but this entry could not be found in the STAR file. "
                "The `selection_filter` must be a dictionary whose "
                "keys are strings in the STAR file and whose values "
                "are functions that take in columns and return boolean "
                "masks."
            )
    # Select particles using mask
    starfile_data["particles"] = particle_data[boolean_mask]

    return starfile_data


#
# STAR file reading
#
def _make_pytrees_from_starfile(
    particle_data,
    optics_data,
    broadcasts_image_config,
    loads_envelope,
    pad_options,
    rotation_convention,
) -> tuple[BasicImageConfig, ContrastTransferTheory, EulerAnglePose]:
    float_dtype = jax.dtypes.canonicalize_dtype(float)
    # Load CTF parameters. First from particle data
    defocus_in_angstroms = (
        np.asarray(particle_data["rlnDefocusU"], dtype=float_dtype)
        + np.asarray(particle_data["rlnDefocusV"], dtype=float_dtype)
    ) / 2
    astigmatism_in_angstroms = np.asarray(
        particle_data["rlnDefocusU"], dtype=float_dtype
    ) - np.asarray(particle_data["rlnDefocusV"], dtype=float_dtype)
    astigmatism_angle = np.asarray(particle_data["rlnDefocusAngle"], dtype=float_dtype)
    phase_shift = np.asarray(particle_data["rlnPhaseShift"], dtype=float_dtype)
    # Then from optics data
    batch_shape = (
        () if defocus_in_angstroms.ndim == 0 else (defocus_in_angstroms.shape[0],)
    )
    spherical_aberration_in_mm = np.full(
        batch_shape, np.asarray(optics_data["rlnSphericalAberration"], dtype=float_dtype)
    )
    amplitude_contrast_ratio = np.full(
        batch_shape, np.asarray(optics_data["rlnAmplitudeContrast"], dtype=float_dtype)
    )
    ctf_params = (
        defocus_in_angstroms,
        astigmatism_in_angstroms,
        astigmatism_angle,
        spherical_aberration_in_mm,
        amplitude_contrast_ratio,
        phase_shift,
    )
    # Envelope parameters
    if loads_envelope:
        b_factor, scale_factor = (
            (
                np.asarray(particle_data["rlnCtfBfactor"], dtype=float_dtype)
                if "rlnCtfBfactor" in particle_data.keys()
                else None
            ),
            (
                np.asarray(particle_data["rlnCtfScalefactor"], dtype=float_dtype)
                if "rlnCtfScalefactor" in particle_data.keys()
                else None
            ),
        )
    else:
        b_factor, scale_factor = None, None
    # Image config parameters
    pixel_size = np.asarray(optics_data["rlnImagePixelSize"], dtype=float_dtype)
    voltage_in_kilovolts = np.asarray(optics_data["rlnVoltage"], dtype=float_dtype)
    if broadcasts_image_config and len(batch_shape) > 0:
        pixel_size = np.full(batch_shape, pixel_size)
        voltage_in_kilovolts = np.full(batch_shape, voltage_in_kilovolts)
    # Pose parameters. Values for the pose are optional,
    # so look to see if each key is present
    particle_keys = particle_data.keys()
    # Read the pose. first, xy offsets
    rln_origin_x_angst = (
        particle_data["rlnOriginXAngst"] if "rlnOriginXAngst" in particle_keys else 0.0
    )
    rln_origin_y_angst = (
        particle_data["rlnOriginYAngst"] if "rlnOriginYAngst" in particle_keys else 0.0
    )
    # ... rot angle
    rln_angle_rot = (
        particle_data["rlnAngleRot"] if "rlnAngleRot" in particle_keys else 0.0
    )
    # ... tilt angle
    if "rlnAngleTilt" in particle_keys:
        rln_angle_tilt = particle_data["rlnAngleTilt"]
    elif "rlnAngleTiltPrior" in particle_keys:  # support for helices
        rln_angle_tilt = particle_data["rlnAngleTiltPrior"]
    else:
        rln_angle_tilt = 0.0
    # ... psi angle
    if "rlnAnglePsi" in particle_keys:
        # Relion uses -999.0 as a placeholder for an un-estimated in-plane
        # rotation
        if isinstance(particle_data["rlnAnglePsi"], pd.Series):
            # ... check if all values are equal to -999.0. If so, just
            # replace the whole pandas.Series with 0.0
            if (
                particle_data["rlnAnglePsi"].nunique() == 1
                and particle_data["rlnAnglePsi"].iloc[0] == -999.0
            ):
                rln_angle_psi = 0.0
            # ... otherwise, replace -999.0 values with 0.0
            else:
                rln_angle_psi = particle_data["rlnAnglePsi"].where(
                    lambda x: x != -999.0, 0.0
                )
        else:
            # ... if the column is just equal to a float, then
            # directly check if it is equal to -999.0
            rln_angle_psi = (
                0.0
                if particle_data["rlnAnglePsi"] == -999.0
                else particle_data["rlnAnglePsi"]
            )
    elif "rlnAnglePsiPrior" in particle_keys:  # support for helices
        rln_angle_psi = particle_data["rlnAnglePsiPrior"]
    else:
        rln_angle_psi = 0.0
    # Now, flip the sign of the translations and transpose rotations.
    # RELION's convention thinks about the translation as "undoing" the translation
    # and rotation in the image
    maybe_make_full = lambda param: (
        np.full(batch_shape, param)
        if len(batch_shape) > 0 and param.shape == ()
        else param
    )
    pose_params = tuple(
        maybe_make_full(x)
        for x in (
            -np.asarray(rln_origin_x_angst, dtype=float_dtype),
            -np.asarray(rln_origin_y_angst, dtype=float_dtype),
            -np.asarray(rln_angle_rot, dtype=float_dtype),
            -np.asarray(rln_angle_tilt, dtype=float_dtype),
            -np.asarray(rln_angle_psi, dtype=float_dtype),
        )
    )
    # Now, create cryojax objects. Do this on the CPU
    cpu_device = jax.devices(backend="cpu")[0]
    with jax.default_device(cpu_device):
        # First, create the `BasicImageConfig`
        image_size = int(optics_data["rlnImageSize"])
        image_shape = (image_size, image_size)
        image_config = _make_config(
            image_shape, pixel_size, voltage_in_kilovolts, pad_options
        )
        # ... now the `ContrastTransferTheory`
        envelope = (
            _make_envelope_function(scale_factor, b_factor) if loads_envelope else None
        )
        transfer_theory_params = (*ctf_params, envelope)
        transfer_theory = _make_transfer_theory(*transfer_theory_params)  # type: ignore
        # ... finally the `EulerAnglePose`
        pose = _make_pose(*pose_params)
        if rotation_convention == "frame":
            pose = _invert_rotation(pose)
    # Now, convert arrays to numpy in case the user wishes to do preprocessing
    pytree_dynamic, pytree_static = eqx.partition(
        (image_config, transfer_theory, pose), eqx.is_array
    )
    pytree_dynamic = jax.tree.map(lambda x: np.asarray(x), pytree_dynamic)
    image_config, transfer_theory, pose = eqx.combine(pytree_dynamic, pytree_static)

    return image_config, transfer_theory, pose


def _make_config(
    image_shape,
    pixel_size,
    voltage_in_kilovolts,
    pad_options,
):
    return eqx.tree_at(
        lambda x: (x.pixel_size, x.voltage_in_kilovolts),
        BasicImageConfig(image_shape, 1.0, 1.0, pad_options=pad_options),
        (pixel_size, voltage_in_kilovolts),
    )


def _make_pose(offset_x, offset_y, phi, theta, psi):
    _make_fn = lambda _x, _y, _phi, _theta, _psi: EulerAnglePose(
        _x, _y, _phi, _theta, _psi
    )
    if offset_x.ndim == 1:
        _make_fn = eqx.filter_vmap(_make_fn)
    return _make_fn(offset_x, offset_y, phi, theta, psi)


def _make_envelope_function(amp, b_factor):
    if b_factor is None and amp is None:
        warnings.warn(
            "`loads_envelope` was set to True, but no envelope parameters were found. "
            "Setting envelope as None. "
            "Make sure your starfile is correctly formatted or set "
            "`loads_envelope=False`."
        )
        return None

    elif b_factor is None and amp is not None:
        return eqx.tree_at(lambda x: x.value, FourierConstant(1.0), amp)
    else:
        if amp is None:
            amp = np.asarray(1.0) if b_factor.ndim == 0 else np.ones_like(b_factor)
        return eqx.tree_at(
            lambda x: (x.amplitude, x.b_factor),
            FourierGaussian(1.0, 1.0),
            (amp, b_factor),
        )


def _make_transfer_theory(defocus, astig, angle, sph, ac, ps, env=None):
    ctf = eqx.tree_at(
        lambda x: (
            x.defocus_in_angstroms,
            x.astigmatism_in_angstroms,
            x.astigmatism_angle,
            x.spherical_aberration_in_mm,
        ),
        AstigmaticCTF(),
        (defocus, astig, angle, sph),
    )
    transfer_theory = ContrastTransferTheory(
        ctf, envelope=env, amplitude_contrast_ratio=0.1, phase_shift=0.0
    )

    return eqx.tree_at(
        lambda x: (x.amplitude_contrast_ratio, x.phase_shift), transfer_theory, (ac, ps)
    )


def _invert_rotation(pose: EulerAnglePose) -> EulerAnglePose:
    negate_angle = lambda angle: ((-angle + 180) % 360) - 180
    return eqx.tree_at(
        lambda x: (x.phi_angle, x.theta_angle, x.psi_angle),
        pose,
        (
            negate_angle(pose.psi_angle),
            negate_angle(pose.theta_angle),
            negate_angle(pose.phi_angle),
        ),
    )


def _load_image_stack_from_mrc(
    shape: tuple[int, int],
    particle_dataframe_at_index: pd.DataFrame,
    path_to_relion_project: str | pathlib.Path,
) -> Float[np.ndarray, "... y_dim x_dim"]:
    # Load particle image stack rlnImageName
    rln_image_names = particle_dataframe_at_index["rlnImageName"].reset_index(drop=True)
    if rln_image_names.convert_dtypes().dtype != "string":
        raise TypeError(
            "The 'rlnImageName' column was expected to be type string. "
            f"Instead, found type {rln_image_names.dtype}."
        )
    # Split the pandas.Series into two: one for the image index
    # and another for the filename
    split = rln_image_names.str.split("@").str
    # ... relion convention starts indexing at 1, not 0
    mrc_index, filenames = split[0], split[1]
    if filenames.isnull().any() or mrc_index.isnull().any():
        raise ValueError(
            "Could not parse filenames from the 'rlnImageName' "
            "column. Check to make sure these are the correct format, "
            "for example '00000@path/to/image-00000.mrcs'."
        )
    mrc_index = mrc_index.astype(int) - 1
    # Allocate memory for stack
    n_images = len(filenames)
    image_stack = np.empty((n_images, *shape), dtype=float)
    # Loop over filenames to fill stack
    unique_filenames = filenames.unique()
    for filename in unique_filenames:
        # Get the MRC indices
        mrc_index_at_filename = mrc_index[filename == filenames]
        particle_index_at_filename = mrc_index_at_filename.index
        path_to_filename = pathlib.Path(path_to_relion_project, filename)
        with mrcfile.mmap(path_to_filename, mode="r", permissive=True) as mrc:
            mrc_data = np.asarray(mrc.data)
            mrc_ndim = mrc_data.ndim
            mrc_shape = mrc_data.shape if mrc_ndim == 2 else mrc_data.shape[1:]

            if shape != mrc_shape:
                raise ValueError(
                    f"The shape of the MRC with filename {filename} "
                    "was found to not have the same shape loaded from "
                    "the 'rlnImageSize'. Check your MRC files and also "
                    "the STAR file optics group formatting."
                )
            image_stack[particle_index_at_filename] = (
                mrc_data if mrc_ndim == 2 else mrc_data[mrc_index_at_filename]
            )

    return image_stack


def _validate_dataset_index(cls, index, n_rows):
    index_error_msg = lambda idx: (
        f"The index at which the `{cls.__name__}` was accessed was out of bounds! "
        f"The number of rows in the dataset is {n_rows}, but you tried to "
        f"access the index {idx}."
    )
    # ... pandas has bad error messages for its indexing
    if isinstance(index, (int, np.integer)):  # type: ignore
        if index > n_rows - 1:
            raise IndexError(index_error_msg(index))
    elif isinstance(index, slice):
        if index.start is not None and index.start > n_rows - 1:
            raise IndexError(index_error_msg(index.start))
    elif isinstance(index, np.ndarray):
        if index.size == 0:
            raise IndexError(
                "Found that the index passed to the dataset "
                "was an empty numpy array. Please pass a "
                "supported index."
            )
    else:
        raise IndexError(
            f"Indexing with the type {type(index)} is not supported by "
            f"`{cls.__name__}`. Indexing by integers is supported, one-dimensional "
            "fancy indexing is supported, and numpy-array indexing is supported. "
            "For example, like `particle = particle_dataset[0]`, "
            "`particle_stack = particle_dataset[0:5]`, "
            "or `particle_stack = dataset[np.array([1, 4, 3, 2])]`."
        )


def _validate_starfile_data(starfile_data: dict[str, pd.DataFrame]):
    if "particles" not in starfile_data.keys():
        raise ValueError("Missing key 'particles' in `starfile.read` output.")
    else:
        required_particle_keys, _ = zip(*RELION_REQUIRED_PARTICLE_ENTRIES)
        if not set(required_particle_keys).issubset(
            set(starfile_data["particles"].keys())
        ):
            raise ValueError(
                "Missing required keys in starfile 'particles' group. "
                f"Required keys are {required_particle_keys}."
            )
    if "optics" not in starfile_data.keys():
        raise ValueError("Missing key 'optics' in `starfile.read` output.")
    else:
        required_optics_keys, _ = zip(*RELION_REQUIRED_OPTICS_ENTRIES)
        if not set(required_optics_keys).issubset(set(starfile_data["optics"].keys())):
            raise ValueError(
                "Missing required keys in starfile 'optics' group. "
                f"Required keys are {required_optics_keys}."
            )


def _validate_rln_image_name_exists(particle_data, index):
    if "rlnImageName" not in particle_data.columns:
        raise OSError(
            "Tried to read STAR file for "
            f"`RelionParticleDataset` index = {index}, "
            "but no entry found for 'rlnImageName'."
        )


#
# Working with pytrees for I/O
#
def _unpack_particle_stack_dict(
    value: _ParticleStackLike,
) -> tuple[Float[NDArrayLike, "... y_dim x_dim"], _ParticleParameterLike | None]:
    if "images" in value:
        images = value["images"]
    else:
        raise ValueError(
            "When passing dictionary `foo` as `dataset.append(foo)` or "
            "`dataset[index] = foo`, `foo` must have key `images`."
        )
    if "parameters" in value:
        parameters = value["parameters"]
    else:
        parameters = None

    return images, parameters


#
# STAR file writing. First, functions for writing parameters
#
def _validate_parameters(parameters: _ParticleParameterLike, force_keys: bool = False):
    if force_keys:
        if not {"image_config", "transfer_theory", "pose"}.issubset(parameters):
            raise ValueError(
                "When passing dictionary `foo` as `parameter_file.append(foo)` "
                "`foo` must have keys 'pose', 'transfer_theory', and 'image_config'."
            )
    if "image_config" in parameters:
        if not isinstance(parameters["image_config"], BasicImageConfig):
            raise TypeError(
                "Found that dict key 'image_config' was "
                "not type `cryojax.simulator.BasicImageConfig`. "
                f"Instead, it was type "
                f"{type(parameters['image_config']).__name__}."
            )
    if "transfer_theory" in parameters:
        if not isinstance(parameters["transfer_theory"], ContrastTransferTheory):
            raise TypeError(
                "Found that dict key 'transfer_theory' was "
                "not type `cryojax.simulator.ContrastTransferTheory`. "
                f"Instead, it was type "
                f"{type(parameters['transfer_theory']).__name__}."
            )
    if "pose" in parameters:
        if not isinstance(parameters["pose"], EulerAnglePose):
            raise TypeError(
                "Found that dict key 'pose' was "
                "not type `cryojax.simulator.EulerAnglePose`. "
                f"Instead, it was type "
                f"{type(parameters['pose']).__name__}."
            )


def _parameters_to_optics_data(
    parameters: _ParticleParameterLike, optics_group_index: int
) -> pd.DataFrame:
    if {"image_config", "transfer_theory"}.issubset(parameters):
        shape = parameters["image_config"].shape
        if shape[0] == shape[1]:
            dim = shape[0]
        else:
            raise ValueError(
                "When adding optics group to STAR file, found "
                "non-square shape in `image_config.shape`. Only "
                "square shapes are supported."
            )
        pixel_size = parameters["image_config"].pixel_size
        voltage_in_kilovolts = parameters["image_config"].voltage_in_kilovolts
        amplitude_contrast_ratio = parameters["transfer_theory"].amplitude_contrast_ratio
        if isinstance(parameters["transfer_theory"].ctf, AstigmaticCTF):
            spherical_aberration_in_mm = getattr(
                parameters["transfer_theory"].ctf, "spherical_aberration_in_mm"
            )
        else:
            raise ValueError(
                "When adding optics group or particle to STAR file, "
                "`transfer_theory.ctf` must be type "
                "`AstigmaticCTF`. Instead, got type "
                f"{type(parameters['transfer_theory'].ctf).__name__}."
            )
        optics_group_dict = {
            "rlnOpticsGroup": optics_group_index,
            "rlnImageSize": dim,
            "rlnImagePixelSize": pixel_size,
            "rlnVoltage": voltage_in_kilovolts,
            "rlnSphericalAberration": spherical_aberration_in_mm,
            "rlnAmplitudeContrast": amplitude_contrast_ratio,
        }
        for k, v in optics_group_dict.items():
            if isinstance(v, Array | np.ndarray):
                arr = np.atleast_1d(np.asarray(v))
                if arr.size > 1:
                    if np.unique(arr).size > 1:
                        raise ValueError(
                            "Tried to fill a RELION optics group entry with an array "
                            "that has multiple unique values. Optics group compatible "
                            "arrays such as `BasicImageConfig.pixel_size` "
                            "must be either scalars or arrays all with the same value. "
                            f"Error occurred when filling '{k}' with array {v}."
                        )
                optics_group_dict[k] = arr.ravel()[0, None]
            else:
                optics_group_dict[k] = [v]
        optics_data = pd.DataFrame.from_dict(optics_group_dict)

        return optics_data
    else:
        raise ValueError(
            "Tried to add optics group to the STAR file, but "
            "parameter dictionary did not include 'image_config' "
            "and 'transfer_theory' keys. If you are setting parameters "
            "`parameter_file[index] = dict(pose=pose)` make sure "
            "`parameter_file.updates_optics_group = False`."
        )


def _parameters_to_particle_data(
    parameters: _ParticleParameterLike,
    optics_group_index: int | None = None,
) -> pd.DataFrame:
    particles_dict = {}
    if "pose" in parameters:
        # Now, pose parameters
        pose = parameters["pose"]
        if not isinstance(pose, EulerAnglePose):
            raise ValueError(
                "When adding particle to STAR file, "
                "`pose` must be type "
                "`EulerAnglePose`. Instead, got type "
                f"{type(pose).__name__}."
            )
        if pose.offset_in_angstroms.ndim == 2:
            particles_dict["rlnOriginXAngst"] = -pose.offset_in_angstroms[:, 0]
            particles_dict["rlnOriginYAngst"] = -pose.offset_in_angstroms[:, 1]
        elif pose.offset_in_angstroms.ndim == 1:
            particles_dict["rlnOriginXAngst"] = -pose.offset_in_angstroms[0]
            particles_dict["rlnOriginYAngst"] = -pose.offset_in_angstroms[1]
        else:
            raise RuntimeError(
                "Internal `cryojax` error when loading translations to STAR file."
            )
        particles_dict["rlnAngleRot"] = -pose.phi_angle
        particles_dict["rlnAngleTilt"] = -pose.theta_angle
        particles_dict["rlnAnglePsi"] = -pose.psi_angle
        # Now, broadcast parameters to same dimension
        n_particles = pose.offset_x_in_angstroms.size
        for k, v in particles_dict.items():
            if v.shape == ():
                particles_dict[k] = np.full((n_particles,), np.asarray(v))
            elif v.size == n_particles:
                particles_dict[k] = v.ravel()
            else:
                raise ValueError(
                    "Found inconsistent number of particles "
                    "when adding particle to STAR file. "
                    "When running `parameter_file[index] = foo` "
                    "or `parameter_file.append(foo)`, make sure `foo` "
                    "has arrays that are scalars or all have the same "
                    "number of dimensions."
                )
    else:
        raise ValueError(
            "Tried to modify parameters in STAR file, but "
            "parameter dictionary did not include 'pose' "
            "key. If you are setting parameters "
            "`parameter_file[index] = foo` make sure "
            "`foo` is, for example, `foo = dict(pose=EulerAnglePose(...))`."
        )
    # Fill CTF parameters
    if "transfer_theory" in parameters:
        transfer_theory = parameters["transfer_theory"]
        if isinstance(transfer_theory.ctf, AstigmaticCTF):
            if pose.offset_z_in_angstroms is None:
                defocus_offset = 0.0
            else:
                defocus_offset = pose.offset_z_in_angstroms
            particles_dict["rlnDefocusU"] = (
                transfer_theory.ctf.defocus_in_angstroms
                + defocus_offset
                + transfer_theory.ctf.astigmatism_in_angstroms / 2
            )
            particles_dict["rlnDefocusV"] = (
                transfer_theory.ctf.defocus_in_angstroms
                + defocus_offset
                - transfer_theory.ctf.astigmatism_in_angstroms / 2
            )
            particles_dict["rlnDefocusAngle"] = transfer_theory.ctf.astigmatism_angle
        else:
            raise ValueError(
                "When adding particle to STAR file, "
                "`transfer_theory.ctf` must be type "
                "`AstigmaticCTF`. Instead, got type "
                f"{type(parameters['transfer_theory'].ctf).__name__}."
            )

        if isinstance(transfer_theory.envelope, FourierGaussian):
            particles_dict["rlnCtfBfactor"] = transfer_theory.envelope.b_factor
            particles_dict["rlnCtfScalefactor"] = transfer_theory.envelope.amplitude
        elif isinstance(transfer_theory.envelope, FourierConstant):
            particles_dict["rlnCtfScalefactor"] = transfer_theory.envelope.value
        elif transfer_theory.envelope is None:
            pass
        else:
            raise ValueError(
                "When adding particle to STAR file, "
                "`transfer_theory.envelope` must be "
                "type `cryojax.ndimage.FourierGaussian` "
                "or `cryojax.ndimage.FourierConstant`, or `None`. Got "
                f"{type(transfer_theory.envelope).__name__}."
            )
        particles_dict["rlnPhaseShift"] = transfer_theory.phase_shift
    # Now, miscellaneous parameters
    if optics_group_index is not None:
        particles_dict["rlnOpticsGroup"] = np.full(
            (n_particles,), optics_group_index, dtype=int
        )
    # Make sure parameters are numpy arrays
    particles_dict = jax.tree.map(
        lambda x: np.asarray(x) if isinstance(x, Array) else x, particles_dict
    )
    particle_data = pd.DataFrame.from_dict(particles_dict)
    # Finally, see if the particle parameters has metadata and if so,
    # add this
    if "metadata" in parameters:
        if parameters["metadata"] is not None:
            metadata = cast(pd.DataFrame, parameters["metadata"])
            if n_particles != metadata.index.size:
                raise ValueError(
                    "When adding custom metadata to STAR file "
                    "with `parameter_file[index] = foo` or `parameter_file.append(foo)`, "
                    "found the number of particles "
                    "in `foo['metadata']` was inconsistent with the "
                    "number of particles in `foo['pose']`."
                )
            # Add metadata to dataframe
            particle_data = pd.concat(
                [particle_data, metadata], axis="columns", verify_integrity=True
            )
    return particle_data


def _make_optics_group_index(optics_data: pd.DataFrame) -> int:
    optics_group_indices = np.asarray(optics_data["rlnOpticsGroup"], dtype=int)
    last_optics_group_index = (
        0 if optics_group_indices.size == 0 else int(optics_group_indices[-1])
    )
    return last_optics_group_index + 1


def _parse_optics_group_index(particle_data_at_index: pd.DataFrame | pd.Series) -> int:
    # ... read optics data
    optics_group_indices = np.unique(
        np.atleast_1d(np.asarray(particle_data_at_index["rlnOpticsGroup"]))
    )
    if optics_group_indices.size > 1:
        raise NotImplementedError(
            "Tried to read multiple particles at once that belong "
            "to different optics groups, but this is not yet "
            "implemented. In the meantime, try reading one particle "
            "at a time."
        )
    optics_group_index = optics_group_indices[0]

    return int(optics_group_index)


def _get_optics_group_from_index(
    optics_data: pd.DataFrame, optics_group_index: int
) -> pd.Series:
    return optics_data[optics_data["rlnOpticsGroup"] == optics_group_index].iloc[0]


def _get_optics_group_from_particle_data(
    particle_data_at_index: pd.DataFrame | pd.Series, optics_data: pd.DataFrame
) -> pd.Series:
    optics_group_index = _parse_optics_group_index(particle_data_at_index)
    return _get_optics_group_from_index(optics_data, optics_group_index)


#
# Now, functions for writing image files
#
def _dict_to_mrcfile_settings(d: dict[str, Any]) -> _MrcfileSettings:
    prefix = d["prefix"] if "prefix" in d else ""
    output_folder = d["output_folder"] if "output_folder" in d else ""
    delimiter = d["delimiter"] if "delimiter" in d else "_"
    n_characters = d["n_characters"] if "n_characters" in d else 6
    overwrite = d["overwrite"] if "overwrite" in d else False
    compression = d["compression"] if "compression" in d else None
    return _MrcfileSettings(
        prefix=prefix,
        output_folder=output_folder,
        delimiter=delimiter,
        n_characters=n_characters,
        overwrite=overwrite,
        compression=compression,
    )


def _index_to_array(indices: slice | int | np.ndarray, size: int) -> np.ndarray:
    if isinstance(indices, slice):
        return np.asarray(range(*indices.indices(size)))
    else:
        return np.asarray(indices, dtype=int)


def _make_image_filename(
    index: Int[np.ndarray, " _"],
    particle_data: pd.DataFrame,
    n_particles: int,
    mrcfile_settings: _MrcfileSettings,
    path_to_relion_project: pathlib.Path,
) -> tuple[pathlib.Path, list[str]]:
    # Get the file number for this MRC file
    if n_particles == 0:
        file_number = 0
    else:
        last_index = index[0] - 1
        if last_index == -1:
            file_number = 0
        else:
            last_filename = particle_data["rlnImageName"].iloc[last_index].split("@")[1]
            if pd.isna(last_filename):
                raise OSError(
                    "Tried to assign a number to the MRC file while writing "
                    "images, but could not grab the previous file number at "
                    f"index {int(last_index)}. At this index, found that the "
                    "filename was NaN."
                )
            else:
                file_number = _parse_filename_for_number(last_filename) + 1
    # Unpack settings
    prefix = mrcfile_settings["prefix"]
    output_folder = mrcfile_settings["output_folder"]
    delimiter = mrcfile_settings["delimiter"]
    n_characters = mrcfile_settings["n_characters"]
    # Generate filename
    file_number_fmt = _format_number_for_filename(file_number, n_characters=n_characters)
    if prefix == "":
        relative_path_to_filename = str(
            pathlib.Path(output_folder, file_number_fmt + ".mrcs")
        )
    else:
        relative_path_to_filename = str(
            pathlib.Path(output_folder, prefix + delimiter + file_number_fmt + ".mrcs")
        )
    # Finally, generate the 'rln_image_name' column, which includes the particle index
    rln_image_names = [
        _format_number_for_filename(int(i + 1), n_characters)
        + "@"
        + relative_path_to_filename
        for i in range(index.size)
    ]
    # Finally, the path to the filename
    path_to_filename = pathlib.Path(path_to_relion_project, relative_path_to_filename)

    return path_to_filename, rln_image_names


def _parse_filename_for_number(filename: str) -> int:
    match = re.search(r"(\d+)\.[^.]+$", filename)
    try:
        file_number = int(match.group(1))  # type: ignore
    except Exception as err:
        raise OSError(
            f"Could not get the file number from file {filename} "
            "Files must be enumerated with the trailing part of the "
            "filename as the file number, like so: '/path/to/file-0000.txt'. "
            f"When extracting the file number and converting it to an integer, "
            f"found error:\n\t{err}"
        )
    return file_number


def _format_number_for_filename(file_number: int, n_characters: int = 6):
    if file_number == 0:
        return "0" * n_characters
    else:
        n_digits = int(np.log10(file_number)) + 1
        return "0" * (n_characters - n_digits) + str(file_number)



import warnings as _warnings


_warnings.warn(
    "Submodule `cryojax.coordinates` is deprecated and "
    "has been moved to the `cryojax.ndimage` namespace. "
    "`cryojax.coordinates` will be removed in cryoJAX 0.6.0.",
    category=FutureWarning,
    stacklevel=2,
)

from ..ndimage._coordinates import (
    cartesian_to_polar as cartesian_to_polar,
    make_1d_coordinate_grid as make_1d_coordinate_grid,
    make_1d_frequency_grid as make_1d_frequency_grid,
    make_coordinate_grid as make_coordinate_grid,
    make_frequency_grid as make_frequency_grid,
    make_frequency_slice as make_frequency_slice,
    make_radial_coordinate_grid as make_radial_coordinate_grid,
    make_radial_frequency_grid as make_radial_frequency_grid,
)



"""
Abstraction of the ice in a cryo-EM image.
"""

from abc import abstractmethod
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from .._internal import error_if_negative
from ..constants import PARKHURST2024_POWER_CONSTANTS
from ..ndimage import (
    AbstractFourierOperator,
    FourierGaussian,
    PeakedFourierGaussian,
    ifftn,
    irfftn,
    rescale_fft,
)
from ._image_config import AbstractImageConfig


class SolventMixturePower(AbstractFourierOperator, strict=True):
    r"""A model for the power of the gaussian random field (GRF) for solvent in cryo-EM.
    This implementation takes inspiration from Parkhurst et al. (2024) and uses its
    implementation as default.

    Parkhurst et al. (2024) models the power as a sum of two gaussians, with
    one gaussian modeling an envelope for the low-resolution decay of the power and
    another gaussian modeling the peak from the solvent at high-resolutions. This
    expression is given by

    .. math::
        P(k) = a_1 \exp(-k^2/(2 s_1^2)) + a_2 \exp(-(k-m)^2/(2 s_2^2)),

    where index `1` is for the envelope and index `2` is for the high-resolution peak.

    More generally, this class models the power as a mixture of any two functions, called
    `envelope_fn` and `peak_fn`.
    """

    envelope_fn: AbstractFourierOperator
    peak_fn: AbstractFourierOperator

    def __init__(
        self,
        envelope_fn: AbstractFourierOperator | None = None,
        peak_fn: AbstractFourierOperator | None = None,
    ):
        if envelope_fn is None:
            self.envelope_fn = FourierGaussian(
                amplitude=PARKHURST2024_POWER_CONSTANTS.a_1,
                b_factor=2 / PARKHURST2024_POWER_CONSTANTS.s_1**2,
            )
        if peak_fn is None:
            self.peak_fn = PeakedFourierGaussian(
                amplitude=PARKHURST2024_POWER_CONSTANTS.a_2,
                b_factor=2 / PARKHURST2024_POWER_CONSTANTS.s_2**2,
                radial_peak=PARKHURST2024_POWER_CONSTANTS.m,
            )

    @override
    def __call__(
        self, frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        mixture_fn = self.envelope_fn + self.peak_fn
        return mixture_fn(frequency_grid_in_angstroms)


class AbstractRandomSolvent2D(eqx.Module, strict=True):
    """Base class for a model of the solvent in cryo-EM."""

    thickness_in_angstroms: eqx.AbstractVar[Float[Array, ""]]

    @abstractmethod
    def sample_solvent_in_plane_potential(
        self,
        rng_key: PRNGKeyArray,
        config: AbstractImageConfig,
        outputs_rfft: bool = True,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
        | Float[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
    ):
        """Sample a stochastic realization of scattering potential due to the ice
        at the exit plane."""
        raise NotImplementedError

    def compute_in_plane_potential(
        self,
        rng_key: PRNGKeyArray,
        fourier_in_plane_potential: (
            Complex[
                Array,
                "{config.padded_y_dim} {config.padded_x_dim//2+1}",
            ]
            | Complex[
                Array,
                "{config.padded_y_dim} {config.padded_x_dim}",
            ]
        ),
        config: AbstractImageConfig,
        input_is_rfft: bool = True,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
        | Float[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
    ):
        """Compute the combined spectrum of the ice and the specimen."""
        # Sample the realization of the phase due to the ice.
        # TODO: this function will also handle masking from a model for the
        # solvent shell.
        fourier_in_plane_potential_of_solvent = self.sample_solvent_in_plane_potential(
            rng_key,
            config,
            outputs_rfft=input_is_rfft,
            outputs_real_space=False,
        )
        fourier_in_plane_potential = (
            fourier_in_plane_potential_of_solvent + fourier_in_plane_potential
        )

        if outputs_real_space:
            if input_is_rfft:
                return irfftn(
                    fourier_in_plane_potential,
                    s=config.padded_shape,
                )
            else:
                return ifftn(fourier_in_plane_potential, s=config.padded_shape)
        else:
            return fourier_in_plane_potential


class GRFSolvent2D(AbstractRandomSolvent2D, strict=True):
    r"""Solvent modeled as a gaussian random field (GRF)."""

    thickness_in_angstroms: Float[Array, ""]
    molecules_per_angstrom_cubed: float
    total_potential_per_molecule: float
    power_fn: AbstractFourierOperator
    samples_power: bool

    def __init__(
        self,
        thickness_in_angstroms: Float[Array, ""] | float,
        molecules_per_angstrom_cubed: float = 0.0325,
        total_potential_per_molecule: float = 38.214333,
        power_fn: AbstractFourierOperator | None = None,
        samples_power: bool = False,
    ):
        """**Arguments:**

        - `thickness_in_angstroms`:
            The solvent thickness in angstroms.
        - `potential_per_molecule`:
            A dimensional factor that quantifies the total electrostatic
            potential per molecule. This has units of energy times volume,
            which in `cryojax` conventions is measured in Angstroms.
            By default, this is calibrated as the q = 0 component of
            an electron scattering factor for water, computed from
            `cryojax.constants`.
        - `molecules_per_angstrom_cubed`:
            A dimensional factor that quantifies the number of water molecules
            per unit volume. By default, this is calibrated from MD simulations
            of water.
        - `power_fn` :
            A function that computes the power spectrum of the solvent, up to the
            final mean and standard deviation of the gaussian random field.
        - `samples_power`:
            If `True`, the power spectrum of the sampled GRF is stochastic. If `False`,
            the power is deterministic (i.e. the GRF is generated by sampling
            uniform phases).
        """
        self.power_fn = power_fn or SolventMixturePower()
        self.samples_power = samples_power
        self.thickness_in_angstroms = error_if_negative(
            jnp.asarray(thickness_in_angstroms, dtype=float)
        )
        self.molecules_per_angstrom_cubed = molecules_per_angstrom_cubed
        self.total_potential_per_molecule = total_potential_per_molecule

    @override
    def sample_solvent_in_plane_potential(
        self,
        rng_key: PRNGKeyArray,
        config: AbstractImageConfig,
        outputs_rfft: bool = True,
        outputs_real_space: bool = False,
    ) -> (
        Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim//2+1}",
        ]
        | Complex[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
        | Float[
            Array,
            "{config.padded_y_dim} {config.padded_x_dim}",
        ]
    ):
        """Sample a realization of the ice integrated potential
        as a gaussian random field.
        """
        n_pixels = config.padded_n_pixels
        if outputs_real_space:
            frequency_grid_in_angstroms = config.get_frequency_grid(
                padding=True, physical=True
            )
        else:
            if outputs_rfft:
                frequency_grid_in_angstroms = config.get_frequency_grid(
                    padding=True, physical=True
                )
            else:
                frequency_grid_in_angstroms = config.get_frequency_grid(
                    padding=True, physical=True, full=True
                )
        # Compute standard deviation, scaling up by the variance by the number
        # of pixels to make the realization independent pixel-independent in real-space.
        std = jnp.sqrt(n_pixels * self.power_fn(frequency_grid_in_angstroms))
        solvent_grf = std * jr.normal(
            rng_key,
            shape=frequency_grid_in_angstroms.shape[0:-1],
            dtype=complex,
        ).at[0, 0].set(0.0)
        # Scale to desired mean and standard deviation
        molecules_per_angstrom_squared = (
            self.thickness_in_angstroms * self.molecules_per_angstrom_cubed
        )
        mean = self.total_potential_per_molecule * molecules_per_angstrom_squared
        std = 1.0  # TODO: set standard deviation
        fourier_in_plane_potential_of_solvent = rescale_fft(
            solvent_grf,
            mean=mean,
            std=std,
            input_is_rfft=outputs_rfft,
            shape_in_real_space=config.padded_shape,
        )
        if outputs_real_space:
            return irfftn(
                fourier_in_plane_potential_of_solvent,
                s=config.padded_shape,
            )
        else:
            return fourier_in_plane_potential_of_solvent



"""Implementation of an `AbstractFourierOperator`. Put simply, these are
functions commonly applied to images in fourier space.

Opposed to a `AbstractFilter`, a `AbstractFourierOperator` is computed at
runtime---not upon initialization. `AbstractFourierOperators` also do not
have a rule for how they should be applied to images and can be composed
with other operators.

These classes are modified from the library `tinygp`.
"""

import functools
import operator
from abc import abstractmethod
from collections.abc import Callable
from typing import Any
from typing_extensions import override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact

from ..._internal import error_if_negative, error_if_not_positive
from ...jax_util import FloatLike, NDArrayLike


class AbstractFourierOperator(eqx.Module, strict=True):
    """
    The base class for all fourier-based operators.

    By convention, operators should be defined to
    be dimensionless (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the `__call__` method.
    """

    @abstractmethod
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError

    def __add__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _SumFourierOperator(self, other)
        return _SumFourierOperator(self, FourierConstant(other))

    def __radd__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _SumFourierOperator(other, self)
        return _SumFourierOperator(FourierConstant(other), self)

    def __sub__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _DiffFourierOperator(self, other)
        return _DiffFourierOperator(self, FourierConstant(other))

    def __rsub__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _DiffFourierOperator(other, self)
        return _DiffFourierOperator(FourierConstant(other), self)

    def __mul__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _ProductFourierOperator(self, other)
        return _ProductFourierOperator(self, FourierConstant(other))

    def __rmul__(self, other) -> "AbstractFourierOperator":
        if isinstance(other, AbstractFourierOperator):
            return _ProductFourierOperator(other, self)
        return _ProductFourierOperator(FourierConstant(other), self)


class _SumFourierOperator(AbstractFourierOperator, strict=True):
    """A helper to represent the sum of two operators."""

    operator1: AbstractFourierOperator
    operator2: AbstractFourierOperator

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(frequency_grid) + self.operator2(frequency_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} + {repr(self.operator2)}"


class _DiffFourierOperator(AbstractFourierOperator, strict=True):
    """A helper to represent the difference of two operators."""

    operator1: AbstractFourierOperator
    operator2: AbstractFourierOperator

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(frequency_grid) - self.operator2(frequency_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} - {repr(self.operator2)}"


class _ProductFourierOperator(AbstractFourierOperator, strict=True):
    """A helper to represent the product of two operators."""

    operator1: AbstractFourierOperator
    operator2: AbstractFourierOperator

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.operator1(frequency_grid) * self.operator2(frequency_grid)

    def __repr__(self):
        return f"{repr(self.operator1)} * {repr(self.operator2)}"


class CustomFourierOperator(AbstractFourierOperator, strict=True):
    """An operator that calls a custom function."""

    fn: Callable[..., Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]]
    args: Any
    kwargs: Any

    def __init__(
        self,
        fn: Callable[
            ..., Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]
        ],
        *args: Any,
        **kwargs: Any,
    ):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        return self.fn(frequency_grid, *self.args, **self.kwargs)


CustomFourierOperator.__init__.__doc__ = """**Arguments:**

- `fn`:
    The `Callable` wrapped into a `AbstractFourierOperator`.
    Has signature `out = fn(frequency_grid, *args, **kwargs)`
- `args`:
    Passed to `fn`.
- `kwargs`:
    Passed to `fn`.
"""


class FourierDC(AbstractFourierOperator, strict=True):
    """This operator returns a constant in the DC component."""

    value: Float[Array, ""]

    def __init__(self, value: FloatLike = 0.0):
        """**Arguments:**

        - `value`: The value of the zero mode.
        """
        self.value = jnp.asarray(value)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        return jnp.zeros(frequency_grid.shape[0:-1]).at[0, 0].set(self.value)


class FourierConstant(AbstractFourierOperator, strict=True):
    """An operator that is a constant."""

    value: Float[Array, "..."]

    def __init__(self, value: float | Float[Array, "..."]):
        """**Arguments:**

        - `value`: The value of the constant
        """
        self.value = jnp.asarray(value)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "..."]:
        del frequency_grid
        return self.value


class FourierExp2D(AbstractFourierOperator, strict=True):
    r"""This operator, in real space, represents a
    function equal to an exponential decay, given by

    .. math::
        g(|r|) = \frac{\kappa}{2 \pi \xi^2} \exp(- |r| / \xi),

    where :math:`|r| = \sqrt{x^2 + y^2}` is a radial coordinate.
    Here, :math:`\xi` has dimensions of length and :math:`g(r)`
    has dimensions of inverse area. The power spectrum from such
    a correlation function (in two-dimensions) is given by its
    Hankel transform pair

    .. math::
        P(|k|) = \frac{\kappa}{2 \pi \xi^3} \frac{1}{(\xi^{-2} + |k|^2)^{3/2}}.

    Here :math:`\kappa` is a scale factor and :math:`\xi` is a length
    scale.
    """

    amplitude: Float[Array, ""]
    length_scale: Float[Array, ""]

    def __init__(
        self,
        amplitude: FloatLike = 1.0,
        length_scale: FloatLike = 1.0,
    ):
        """**Arguments:**

        - `amplitude`: The amplitude of the operator, equal to $\\kappa$
                in the above equation.
        - `length_scale`: The length scale of the operator, equal to $\\xi$
                    in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.length_scale = jnp.asarray(length_scale, dtype=float)

    @override
    def __call__(
        self, frequency_grid: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        k_sqr = jnp.sum(frequency_grid**2, axis=-1)
        scaling = (
            1.0
            / (k_sqr + jnp.divide(1, (error_if_not_positive(self.length_scale)) ** 2))
            ** 1.5
        )
        scaling *= jnp.divide(self.amplitude, 2 * jnp.pi * self.length_scale**3)
        return scaling


class FourierGaussian(AbstractFourierOperator, strict=True):
    r"""This operator represents a simple gaussian.
    Specifically, this is

    .. math::
        P(k) = \kappa \exp(- \beta k^2 / 4),

    where :math:`k^2 = k_x^2 + k_y^2` is the length of the
    wave vector. Here, :math:`\beta` has dimensions of length
    squared.
    """

    amplitude: Float[Array, ""]
    b_factor: Float[Array, ""]

    def __init__(self, amplitude: FloatLike = 1.0, b_factor: FloatLike = 1.0):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `b_factor`:
            The B-factor of the gaussian, equal to $\\beta$
            in the above equation.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.b_factor = jnp.asarray(b_factor, dtype=float)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k_sqr = jnp.sum(frequency_grid**2, axis=-1)
        gaussian = self.amplitude * jnp.exp(
            -0.25 * error_if_not_positive(self.b_factor) * k_sqr
        )

        return gaussian


class PeakedFourierGaussian(AbstractFourierOperator, strict=True):
    r"""This operator represents a gaussian with a peak
    at a given frequency shell.
    """

    amplitude: Float[Array, ""]
    b_factor: Float[Array, ""]
    radial_peak: Float[Array, ""]

    def __init__(
        self,
        amplitude: FloatLike = 1.0,
        b_factor: FloatLike = 1.0,
        radial_peak: FloatLike = 0.0,
    ):
        """**Arguments:**

        - `amplitude`:
            The amplitude of the operator, equal to $\\kappa$
            in the above equation.
        - `b_factor`:
            The B-factor of the gaussian, equal to $\\beta$
            in the above equation.
        - `radial_peak`:
            The frequency shell of the gaussian peak.
        """
        self.amplitude = jnp.asarray(amplitude, dtype=float)
        self.b_factor = jnp.asarray(b_factor, dtype=float)
        self.radial_peak = jnp.asarray(radial_peak, dtype=float)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k = jnp.linalg.norm(frequency_grid, axis=-1)
        gaussian = self.amplitude * jnp.exp(
            -0.25
            * error_if_not_positive(self.b_factor)
            * (k - error_if_negative(self.radial_peak)) ** 2
        )
        return gaussian


class FourierSinc(AbstractFourierOperator, strict=True):
    r"""The separable sinc function is the Fourier transform
    of the box function and is commonly used for anti-aliasing
    applications. In 2D, this is

    $$f_{2D}(\vec{q}) = \sinc(q_x w) \sinc(q_y w),$$

    and in 3D this is

    $$f_{3D}(\vec{q}) = \sinc(q_x w) \sinc(q_y w) \sinc(q_z w)},$$

    where $\sinc(x) = \frac{\sin(\pi x)}{\pi x}$,
    $\vec{q} = (q_x, q_y)$ or $\vec{q} = (q_x, q_y, q_z)$ are spatial
    frequency coordinates for 2D and 3D respectively,
    and $w$ is width of the real-space box function.
    """

    box_width: Float[Array, ""]

    def __init__(self, box_width: FloatLike = 1.0):
        """**Arguments:**

        - `box_width`:
            If the inverse fourier transform of this class
            is the rectangular function, its interval is
            `- box_width / 2` to `+ box_width / 2`.
        """
        self.box_width = jnp.asarray(box_width, dtype=float)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        ndim = frequency_grid.ndim - 1
        return functools.reduce(
            operator.mul,
            [jnp.sinc(frequency_grid[..., i] * self.box_width) for i in range(ndim)],
        )


class FourierPhaseShifts(AbstractFourierOperator):
    """Apply a phase shift the Fourier domain."""

    shift: Float[Array, " _"]

    def __init__(self, shift: Float[NDArrayLike, "2"] | Float[NDArrayLike, "3"]):
        """**Arguments:**

        - `shift`:
            The shift to apply in the Fourier domain. The units of this should
            be the inverse of the units of the `frequency_grid` passed at runtime.
        """
        self.shift = jnp.asarray(shift, dtype=float)

    @override
    def __call__(
        self,
        frequency_grid: (
            Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
        ),
    ) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
        ndim = frequency_grid.ndim - 1
        if ndim != self.shift.size:
            raise ValueError(
                "The `frequency_grid` passed to `FourierPhaseShift` had "
                "dimensionality that does not seem to match `FourierPhaseShift.shift`. "
                f"Got that the dimensionality of the grid was `{ndim}`, but the "
                f"shift was an array of size {self.shift.size}"
            )
        return jnp.exp(-1.0j * (2 * jnp.pi * jnp.matmul(frequency_grid, self.shift)))
